from os import getcwd
from os.path import (isfile,
                     join)
from copy import deepcopy
from datetime import date
from openpyxl import load_workbook, Workbook
from openpyxl.styles import Alignment
from win32com.client import Dispatch
from pyodbc import connect as pycon
from sqlite3 import connect as sqcon
from treetopper._constants import (ACCESS_GROUPS_COLS,
                                   ACCESS_STAND_COLS,
                                   ACCESS_TREE_COLS,
                                   SQL_GROUPS_COLS,
                                   SQL_STAND_COLS,
                                   SQL_TREE_COLS,
                                   GROUPS_DEFAULTS,
                                   extension_check)


class FVS(object):
    """The FVS class will create FVS-formatted databases for use in FVS. FVS is a software made by the US Forest Service to
       run models, inventories and simulations on different forest stands. The three types of databases the FVS class creates
       are Microsoft Access, Microsoft Excel and SQLite

       If you would like a blank database to manually input your data, you can simply instantiate the FVS class and call either

       fvs.access_db(filename, directory=yourdirectory (optional), blank_db=True)
       fvs.excel_db(filename, directory=yourdirectory (optional), blank_db=True) OR
       fvs.sqlite_db(filename, directory=yourdirectory (optional), blank_db=True)

       If the directory argument is omitted, the default directory is the current working directory.
       The 'blank_db' argument has to be set to True for blank databases (its default is False)

       If you would like to create the databases from a Stand Class, you can instantiate the FVS class and call the
       fvs.set_stand method. There are seven required arguments, these are data that the FVS software needs for its models, they include

       stand (the Stand Class)
       variant (str) [These are two-letter acronyms for different eco-regions throughout the US]
       forest_code (int) [This is the forest code for the closest national forest to your stand]
       region (int) [This is the region code for the forest service region your stand is in]
       stand_age (int) [The approximate age of the stand]
       site_species (str) [The species acronym for the species being used to measure site index]
       site_index (int) [The site index of the stand]

       kwargs can be used to add additional data to your stand. To see the list of the other stand data call the
       fvs.show_stand_args method, use this console printout to find the index of your desired argument(s),
       then use the list index to set your argument to fill in the kwargs,
       for example fvs.set_stand(... fvs.stand_args[4]=22)

       After setting the stand in the FVS class, you can then create or append your databases by calling the database-specific methods
       listed above"""

    def __init__(self):
        self.stand_args = [i[0] for i in ACCESS_STAND_COLS[1]]
        self.stand = None
        self.stand_fvs = None
        self.tree_fvs = None

    def show_stand_args(self):
        """Shows the optional stand data and their indices within self.stand_args"""
        format_space = 20
        for i, arg in enumerate(self.stand_args):
            print(f'{arg + (" " * (format_space - len(arg)))}\t Index: {i}')

    def set_stand(self, stand, variant: str, forest_code: int, region: int, stand_age: int, site_species: str, site_index: int, **kwargs):
        """Extracts the stand and tree data from the Stand class and updates the stand_fvs and tree_fvs dictionaries which will be used
           to create or update the databases"""
        self.stand = deepcopy(stand)
        self.stand_fvs = {i[0]: None for i in ACCESS_STAND_COLS[1]}
        self.tree_fvs = {}

        if self.stand.inv_date:
            inv_year = self.stand.inv_date.year
        else:
            inv_year = date.today().year

        self.stand_fvs['Stand_ID'] = self.stand.name
        self.stand_fvs['Variant'] = variant
        self.stand_fvs['Inv_Year'] = inv_year
        self.stand_fvs['Groups'] = 'All_Stands'
        self.stand_fvs['Region'] = region
        self.stand_fvs['Forest'] = forest_code
        self.stand_fvs['Age'] = stand_age
        self.stand_fvs['Basal_Area_Factor'] = self.stand.plot_factor
        self.stand_fvs['Brk_DBH'] = 0
        self.stand_fvs['Num_Plots'] = len(self.stand.plots)
        self.stand_fvs['Site_Species'] = site_species
        self.stand_fvs['Site_Index'] = site_index

        row = 1
        for i, plot in enumerate(self.stand.plots):
            pnum = i + 1
            std_plt = f'{self.stand.name}_{pnum}'
            for j, tree in enumerate(plot.trees):
                self.tree_fvs[row] = {i[0]: None for i in ACCESS_TREE_COLS[1]}
                self.tree_fvs[row]['Stand_ID'] = self.stand.name
                self.tree_fvs[row]['StandPlot_ID'] = std_plt
                self.tree_fvs[row]['Plot_ID'] = pnum
                self.tree_fvs[row]['Tree_ID'] = j + 1
                self.tree_fvs[row]['History'] = 1
                self.tree_fvs[row]['Species'] = tree.species
                self.tree_fvs[row]['DBH'] = tree.dbh
                self.tree_fvs[row]['Ht'] = tree.height
                row += 1

        for key in kwargs:
            if key in self.stand_fvs:
                self.stand_fvs[key] = kwargs[key]

    def access_db(self, filename: str, directory: str = None, blank_db: bool = False):
        """Creates or updates an FVS-formatted Microsoft Access Database (.accdb)"""
        db_save = extension_check(filename, '.accdb')

        if directory:
            dir_ = directory
            db_path = join(dir_, db_save)
        else:
            dir_ = getcwd()
            db_path = join(dir_, db_save)

        db_exists = True

        if not isfile(db_path):
            self._create_access_db(db_path)
            db_exists = False

        drive = '{Microsoft Access Driver (*.mdb, *.accdb)}'
        connection_text = r'DRIVER={driver};DBQ={filename};'.format(driver=drive, filename=db_path)
        con = pycon(connection_text)
        cur = con.cursor()

        if not db_exists:
            sql = f"""INSERT INTO FVS_GroupAddFilesAndKeywords (Groups, FVSKeywords) VALUES (?, ?)"""
            cur.execute(sql, [GROUPS_DEFAULTS[0], GROUPS_DEFAULTS[1].format(DB_NAME=db_save)])
            con.commit()

        if not blank_db:
            self._insert_sql(con, cur)
        con.close()

        self._create_loc_file(db_save, dir_)

    def sqlite_db(self, filename: str, directory: str = None, blank_db: bool = False):
        """Creates or updates an FVS-formatted SQLite Database (.db)"""
        db_save = extension_check(filename, '.db')

        if directory:
            db_path = join(directory, db_save)
        else:
            db_path = join(getcwd(), db_save)

        if not isfile(db_path):
            con, cur = self._create_sqlite_db(db_path, db_save)
        else:
            con = sqcon(db_path)
            cur = con.cursor()

        if not blank_db:
            self._insert_sql(con, cur)

        con.close()

    def excel_db(self, filename: str, directory: str = None, blank_db: bool = False):
        """Creates or updates an FVS-formatted Microsoft Excel Database (.xlsx)"""
        db_save = extension_check(filename, '.xlsx')

        if directory:
            db_path = join(directory, db_save)
        else:
            db_path = join(getcwd(), db_save)

        if not isfile(db_path):
            wb = self._create_excel_db(db_save)
        else:
            wb = load_workbook(db_path)

        if not blank_db:
            self._insert_excel(wb)

        wb.save(db_path)

    def _create_access_db(self, db_path:  str):
        """If the file does not exist, this method is called to construct the Access database, used internally"""
        access = Dispatch("Access.Application")
        access_engine = access.DBEngine
        access_workspace = access_engine.Workspaces(0)
        access_language = ';LANGID=0x0409;CP=1252;COUNTRY=0'
        access_db = access_workspace.CreateDatabase(db_path, access_language, 64)

        for table in [ACCESS_GROUPS_COLS, ACCESS_STAND_COLS, ACCESS_TREE_COLS]:
            sql = f"""CREATE TABLE {table[0]} ({', '.join([f'{i[0]} {i[1]}' for i in table[1]])});"""
            access_db.Execute(sql)

    def _create_sqlite_db(self, db_path: str, db_save: str):
        """If the file does not exist, this method is called to construct the SQLite database, used internally"""
        con = sqcon(db_path)
        cur = con.cursor()
        for i, table in enumerate([SQL_GROUPS_COLS, SQL_STAND_COLS, SQL_TREE_COLS]):
            sql = f'CREATE TABLE {table[0]} (' + ', '.join([f'{i[0]} {i[1]}' for i in table[1]]) + ');'
            cur.execute(sql)

        sql = f"""INSERT INTO FVS_GroupAddFilesAndKeywords (Groups, FVSKeywords) VALUES (?, ?)"""
        cur.execute(sql, [GROUPS_DEFAULTS[0], GROUPS_DEFAULTS[1].format(DB_NAME=db_save)])
        con.commit()
        return con, cur

    def _create_excel_db(self, db_save: str):
        """If the file does not exist, this method is called to construct the Excel database, used internally"""
        wb = Workbook()
        for i, table in enumerate([ACCESS_GROUPS_COLS, ACCESS_STAND_COLS, ACCESS_TREE_COLS]):
            ws = wb.create_sheet(table[0])
            for j, col in enumerate(table[1]):
                ws.cell(1, j + 1).value = col[0]
            if i == 0:
                ws.cell(2, 1).value = GROUPS_DEFAULTS[0]
                ws.cell(2, 3).value = GROUPS_DEFAULTS[1].format(DB_NAME=db_save)
                ws.cell(2, 3).alignment = Alignment(wrapText=True)

        for sheet in wb.sheetnames:
            if sheet not in [ACCESS_GROUPS_COLS[0], ACCESS_STAND_COLS[0], ACCESS_TREE_COLS[0]]:
                wb.remove(wb[sheet])
        return wb

    def _insert_sql(self, connection, cursor):
        """Inserts the stand and tree FVS data into the Access or SQLite database"""
        stand_cols = [col for col in self.stand_fvs if self.stand_fvs[col]]
        stand_vals = [self.stand_fvs[col] for col in self.stand_fvs if self.stand_fvs[col]]
        stand_sql = f"""INSERT INTO FVS_StandInit ({', '.join(stand_cols)}) VALUES ({', '.join(['?' for _ in stand_cols])})"""
        cursor.execute(stand_sql, stand_vals)

        for row in self.tree_fvs:
            tree_cols = [col for col in self.tree_fvs[row] if self.tree_fvs[row][col]]
            tree_vals = [self.tree_fvs[row][col] for col in self.tree_fvs[row] if self.tree_fvs[row][col]]
            tree_sql = f"""INSERT INTO FVS_TreeInit ({', '.join(tree_cols)}) VALUES ({', '.join(['?' for _ in tree_cols])})"""

            cursor.execute(tree_sql, tree_vals)
        connection.commit()

    def _insert_excel(self, workbook):
        """Inserts the stand and tree FVS data into the Excel database"""
        stand_keys = list(self.stand_fvs)
        stand_cols = [col for col in self.stand_fvs if self.stand_fvs[col]]
        stand_idxs = [stand_keys.index(col) + 1 for col in stand_cols]
        stand_vals = [self.stand_fvs[col] for col in self.stand_fvs if self.stand_fvs[col]]

        ws = workbook['FVS_StandInit']
        next_row = ws.max_row + 1
        for idx, val in zip(stand_idxs, stand_vals):
            ws.cell(next_row, idx).value = val

        ws = workbook['FVS_TreeInit']
        next_row = ws.max_row + 1
        tree_keys = list(self.tree_fvs[1])
        for row in self.tree_fvs:
            tree_cols = [col for col in self.tree_fvs[row] if self.tree_fvs[row][col]]
            tree_idxs = [tree_keys.index(col) + 1 for col in tree_cols]
            tree_vals = [self.tree_fvs[row][col] for col in self.tree_fvs[row] if self.tree_fvs[row][col]]
            for idx, val in zip(tree_idxs, tree_vals):
                ws.cell(next_row, idx).value = val
            next_row += 1

    def _create_loc_file(self, db_save, directory):
        """In the FVS software (legacy versions), a Suppose.loc file needs to be made or modified with the Access database,
           this file is created in the same directory as the database"""
        loc_path = join(directory, 'Suppose.loc')
        fill_text = f'{loc_path[0]} "{db_save[0:-6]}" {db_save}'

        if isfile(loc_path):
            in_loc = True
            with open(loc_path, 'r') as f:
                lines = [i.strip('\n') for i in f.readlines()]
                if fill_text not in lines:
                    in_loc = False

            if not in_loc:
                with open(loc_path, 'a') as f:
                    f.write('\n' + fill_text)
        else:
            with open(loc_path, 'w') as f:
                f.write(fill_text)







if __name__ == '__main__':
    from treetopper.stand import Stand

    stand = Stand('OK2', 46.94)
    stand.from_csv_full('../example_csv_and_xlsx/Example_CSV_full.csv')
    fvs = FVS()
    fvs.show_stand_args()
    fvs.set_stand(stand, 'PN', 612, 6, 50, 'DF', 120)
    fvs.access_db('spam')




from csv import reader
from openpyxl import load_workbook
from statistics import mean


def import_csv_quick(directory, stand_name):
    hdrs = []
    plots = {}
    with open(directory, 'r') as csv_file:
        csv_read = reader(csv_file)
        next(csv_read)
        for line in csv_read:
            if line[0] == stand_name:
                pnum = int(line[1])
                if pnum not in plots:
                    plots[pnum] = {}
                dbh = float(line[4])
                check_height = line[5]
                if check_height != '':
                    height = int(check_height)
                    hdrs.append(height / (dbh / 12))
                else:
                    height = check_height

                plots[pnum][int(line[2])] = {'species': line[3],
                                             'dbh': dbh,
                                             'height': height,
                                             'pref_log': int(line[6]),
                                             'min_log': int(line[7])}
            else:
                if plots:
                    break

    avg_hdr = mean(hdrs)
    return plots, avg_hdr


def import_csv_full(directory, stand_name):
    hdrs = []
    plots = {}
    with open(directory, 'r') as csv_file:
        csv_read = reader(csv_file)
        next(csv_read)
        for line in csv_read:
            if line[0] == stand_name:
                pnum = int(line[1])
                if pnum not in plots:
                    plots[pnum] = {}
                dbh = float(line[4])
                check_height = line[5]
                if check_height != '':
                    height = int(float(check_height))
                    hdrs.append(height / (dbh / 12))
                else:
                    height = check_height
                tree_num = int(line[2])
                plots[pnum][tree_num] = {'species': line[3],
                                         'dbh': dbh,
                                         'height': height,}
                logs = {}
                stem_height = int(line[6])
                for i in range(7, 40, 4):
                    if line[i] == '':
                        break
                    else:
                        length = int(line[i])
                        stem_height += length + 1
                        grade = line[i + 1].upper()
                        if line[i + 2] != '':
                            defect = int(line[i + 2])
                        else:
                            defect = 0
                        logs[(i // 4) - 1] = [stem_height, length, grade, defect]
                        if i != 40:
                            if line[i + 3] != '':
                                stem_height += int(line[i + 3])
                plots[pnum][tree_num]['logs'] = logs
            else:
                if plots:
                    break

    avg_hdr = mean(hdrs)
    return plots, avg_hdr


def import_excel_quick(directory, stand_name):
    hdrs = []
    plots = {}

    wb = load_workbook(directory, read_only=True, data_only=True)
    ws = wb.active

    table = ws.iter_rows(min_row=2, values_only=True)
    for line in table:
        if line[0] == stand_name:
            pnum = line[1]
            if pnum not in plots:
                plots[pnum] = {}
            dbh = line[4]
            height = line[5]
            if height:
                hdrs.append(height / (dbh / 12))

            plots[pnum][line[2]] = {'species': line[3],
                                    'dbh': dbh,
                                    'height': height,
                                    'pref_log': line[6],
                                    'min_log': line[7]}
        else:
            if plots:
                break

    avg_hdr = mean(hdrs)
    return plots, avg_hdr


def import_excel_full(directory, stand_name):
    hdrs = []
    plots = {}

    wb = load_workbook(directory, read_only=True, data_only=True)
    ws = wb.active

    table = ws.iter_rows(min_row=2, values_only=True)
    for line in table:
        if line[0] == stand_name:
            pnum = line[1]
            if pnum not in plots:
                plots[pnum] = {}
            dbh = line[4]
            height = line[5]
            if height:
                hdrs.append(height / (dbh / 12))
            tree_num = line[2]
            plots[pnum][tree_num] = {'species': line[3],
                                    'dbh': dbh,
                                    'height': height}
            logs = {}
            stem_height = line[6]
            for i in range(7, 40, 4):
                length = line[i]
                if not length:
                    break
                else:
                    stem_height += length + 1
                    grade = line[i+1].upper()
                    if line[i+2]:
                        defect = line[i+2]
                    else:
                        defect = 0
                    logs[(i // 4) - 1] = [stem_height, length, grade, defect]
                    if i != 40:
                        if line[i+3]:
                            stem_height += line[i+3]
            plots[pnum][tree_num]['logs'] = logs

        else:
            if plots:
                break

    avg_hdr = mean(hdrs)
    return plots, avg_hdr



from openpyxl import Workbook
from csv import writer
from os.path import join, isfile, expanduser

METRICS_QUICK = ['Stand', 'Plot Number', 'Tree Number', 'Species', 'DBH', 'Height', 'Preferred Log Length', 'Min Log Length']
METRICS_FULL = ['Stand', 'Plot Number', 'Tree Number', 'Species', 'DBH', 'Height', 'Stump Height',
                'Log 1 Length', 'Log 1 Grade', 'Log 1 Defect %', 'Between Logs Feet',
                'Log 2 Length', 'Log 2 Grade', 'Log 2 Defect %', 'Between Logs Feet',
                'Log 3 Length', 'Log 3 Grade', 'Log 3 Defect %', 'Between Logs Feet',
                'Log 4 Length', 'Log 4 Grade', 'Log 4 Defect %', 'Between Logs Feet',
                'Log 5 Length', 'Log 5 Grade', 'Log 5 Defect %', 'Between Logs Feet',
                'Log 6 Length', 'Log 6 Grade', 'Log 6 Defect %', 'Between Logs Feet',
                'Log 7 Length', 'Log 7 Grade', 'Log 7 Defect %', 'Between Logs Feet',
                'Log 8 Length', 'Log 8 Grade', 'Log 8 Defect %']

DEFAULT_NAME_QUICK = 'stand_data_quick'
DEFAULT_NAME_FULL = 'stand_data_full'

def getdesktoppath():
    return join(expanduser("~"), "desktop")


def create_blank_csv(directory: str = None, full_cruise=False):
    if full_cruise:
        use_name = DEFAULT_NAME_FULL
    else:
        use_name = DEFAULT_NAME_QUICK

    if not directory:
        dir_ = getdesktoppath()
        title = join(dir_, use_name + '.csv')
    else:
        dir_ = directory
        title = join(dir_, use_name + '.csv')

    if isfile(title):
        is_file = True
        count = 1
        while is_file:
            title = join(dir_, use_name + f'_{count}.csv')
            if not isfile(title):
                is_file = False
            else:
                count += 1

    with open(title, 'w', newline='') as csv_file:
        row_write = writer(csv_file)
        if full_cruise:
            row_write.writerow(METRICS_FULL)
        else:
            row_write.writerow(METRICS_QUICK)
    return title



def create_blank_excel(directory: str = None, full_cruise=False):
    if full_cruise:
        use_name = DEFAULT_NAME_FULL
    else:
        use_name = DEFAULT_NAME_QUICK

    if not directory:
        dir_ = getdesktoppath()
        title = join(dir_, use_name + '.xlsx')
    else:
        dir_ = directory
        title = join(dir_, use_name + '.xlsx')

    if isfile(title):
        is_file = True
        count = 1
        while is_file:
            title = join(dir_, use_name + f'_{count}.xlsx')
            if not isfile(title):
                is_file = False
            else:
                count += 1

    wb = Workbook()
    ws = wb.active

    if full_cruise:
        use_metrics = METRICS_FULL
    else:
        use_metrics = METRICS_QUICK

    for i, met in enumerate(use_metrics):
        ws.cell(1, i + 1, met)

    wb.save(title)
    return title


if __name__ == '__main__':
    err = True
    q = None
    while err:
        q = input('\nWould you like a formatted CSV or Excel file (c / e)?: ').upper()
        if q == 'C' or q == 'E':
            err = False
        else:
            print('\nPlease enter either "c" or "e"')

    if q == 'C':
        file_type = 'CSV'
    else:
        file_type = 'Excel'

    err1 = True
    q1 = None
    full = False
    while err1:
        q1 = input('\nWould you like Full Cruise or Quick Cruise sheet (f / q)?: ').upper()
        if q1 == 'F' or q1 == 'Q':
            if q1 == 'F':
                full = True
            err1 = False
        else:
            print('\nPlease enter either "f" or "q"')

    q2 = input(f'\nWould you like to specify a directory?\nElse {file_type} file will be created on desktop\n(y / n): ').upper()
    if q2.upper() == 'Y':
        dir_ = input('\nDirectory path: ')
    else:
        dir_ = None

    try:
        if q == 'E':
            title = create_blank_excel(directory=dir_, full_cruise=full)
        else:
            title = create_blank_csv(directory=dir_, full_cruise=full)

        print(f'\n** Completed {file_type} file creation: "{title}" **')
    except FileNotFoundError:
        print(f'\n-- Original Directory "{dir_}" could not be found, {file_type} file will be created on Desktop --')

        if q == 'E':
            title = create_blank_excel(directory=None, full_cruise=full)
        else:
            title = create_blank_csv(directory=None, full_cruise=full)


        print(f'\n** Completed {file_type} file creation: "{title}" **')






from random import choice, randrange
from time import perf_counter
from treetopper._constants import (ALL_SPECIES_NAMES,
                                   OFFICIAL_GRADES)


SPECIES_RANGE = 3

def timer(func):
    def wrapper(*args, **kwargs):
        now = perf_counter()
        x = func(*args, **kwargs)
        after = perf_counter()
        time_took = round(after-now, 5)
        print(f'{func.__name__} took {time_took} seconds\n')
        return x, time_took
    return wrapper


def calc_log_grade(species, dib, length, next_=False):
    for i, specs in enumerate(OFFICIAL_GRADES[species]):
        if dib >= specs[0] and length >= specs[1]:
            if i != len(OFFICIAL_GRADES[species]) - 1:
                if next_:
                    return OFFICIAL_GRADES[species][i+1][2]
                else:
                    return specs[2]
            else:
                return specs[2]


def generate_random_trees_quick(number_of_trees: int):
    from timber import TimberQuick

    tree_list = []
    for i in range(number_of_trees):
        hdr = randrange(540, 899) / 10
        species = choice([key for key in ALL_SPECIES_NAMES][:SPECIES_RANGE])
        dbh = randrange(69, 455) / 10
        height = int((dbh / 12) * hdr)
        plot_factor = 33.3

        tree_list.append(TimberQuick(*[species, dbh, height, plot_factor]))
    return tree_list, len(tree_list)


def generate_random_trees_full(number_of_trees: int):
    from timber import TimberFull

    tree_list = []
    for i in range(number_of_trees):
        hdr = randrange(540, 899) / 10
        species = choice([key for key in ALL_SPECIES_NAMES][:SPECIES_RANGE])

        dbh = randrange(69, 455) / 10
        height = int((dbh / 12) * hdr)
        plot_factor = 33.3

        tree = TimberFull(*[species, dbh, height, plot_factor])

        x = 1
        max_ = int(height * 0.8)
        defects = [0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 5, 5, 5, 10, 10, 10, 15]
        for i in range(10):
            if x + 41 > max_:
                length = int(((max_ - x - 2) // 2) * 2)
                if length < 2:
                    break
                else:
                    stem_height = max_
                    dib = tree.get_any_dib(stem_height)
                    defect = choice(defects)
                    if defect > 5:
                        grade = calc_log_grade(species, dib, length, next_=True)
                    else:
                        grade = calc_log_grade(species, dib, length)
                    tree.add_log(stem_height, length, grade, defect)
                    break
            else:
                length = 40
                stem_height = x + length + 1
                dib = tree.get_any_dib(stem_height)
                defect = choice(defects)
                if defect > 5:
                    grade = calc_log_grade(species, dib, length, next_=True)
                else:
                    grade = calc_log_grade(species, dib, length)
                tree.add_log(stem_height, length, grade, defect)
                if x + length + 1 < max_:
                    x += length + 1
        tree._calc_volume_and_logs()
        tree_list.append(tree)
    return tree_list, len(tree_list)


def generate_random_plots(number_of_plots: int, full: bool = False):
    from plot import Plot

    tree_count = 0
    plot_list = []

    for i in range(number_of_plots):
        num_trees = randrange(3, 12)
        plot = Plot()
        if full:
            trees, count = generate_random_trees_full(num_trees)
        else:
            trees, count = generate_random_trees_quick(num_trees)

        for j in trees:
            plot.add_tree(j)
        plot_list.append(plot)
        tree_count += count

    return plot_list, tree_count




from treetopper.log import Log
from treetopper._constants import (math,
                                   ALL_TAPERS_DICT,
                                   EQUATION_DICT)



class TimberQuick(object):
    """TimberQuick is a class that will virtually cruise a tree based on it's
       species, DBH, total height and plot factor. For fixed-area plots use the negative inverse of the plot size (1/30th ac = -30),
       for variable-area plots use the Basal Area Factor (BAF) (40 BAF = 40).

       Preferred Log Length and Minimum Log Length are needed
       but set at the default industry standard of 40 foot preferred and 16 foot minimum.

       TimberQuick uses stem-taper equations from Czaplewski, Kozak, or Wensel
       (depending on species) to calculate the DIB (diameter inside bark) at any stem height.

       TimberQuick will instantiate the tree with common individual and per acre metrics based on the input args.

       To cruise the tree, first TimberQuick determines the merchantable DIB of the tree, this is calculated from
       40% of the DIB at a stem height of 17 feet (the FORM height). This is industry standard.

       TimberQuick then correlates that Merch DIB to a merchantable height. The tree is then split up into logs,
       based on this Merch Height with priority given to the preferred log length and if preferred log length
       cannot be achieved, then if the remaining length up to Merch Height is greater than or equal to the minimum log length,
       that final log is added.

       Log metrics are sent to the Log Class, to which their volumes in Board Feet
       (using Scribner Coefficients based on Log Length and top DIB) and Cubic Feet (based on the Two-End Conic Cubic Foot Rule).
       Log grades are determined by species, minimum log lengths and minimum top DIBs set forth by the
       Official Rules for the Log Scaling and Grading Bureaus. Log defect is always 0%

       For inventories, this class is meant to be added to the Plot Class using the Plot Class method of add_tree"""

    def __init__(self, species: str, dbh: float, total_height: int, plot_factor: float,
                 preferred_log_length: int = 40, minimum_log_length: int = 16):
        self.species = str(species).upper()
        self.dbh = float(dbh)
        self.height = int(total_height)
        self.plot_factor = float(plot_factor)
        self.pref_log = int(preferred_log_length)
        self.min_log = int(minimum_log_length)

        self.hdr = self.height / (self.dbh / 12)
        self.ba = self.dbh ** 2 * 0.005454
        self.rd = self.ba / math.sqrt(self.dbh)
        self.coef = ALL_TAPERS_DICT[self.species]
        self.equation = EQUATION_DICT[self.coef[0]]

        self.merch_dib = self._get_merch_dib()
        self.merch_height = self._get_merch_height()

        self.tpa, self.ba_ac, self.rd_ac = 0, 0, 0
        self._get_tpa_ba_ac_rd_ac()

        self.bf = 0
        self.cf = 0
        self.bf_ac = 0
        self.cf_ac = 0
        self.vbar = 0

        self.logs = self._get_volume_and_logs()

    def __getitem__(self, item):
        return self.__dict__[item]

    def get_any_dib(self, stem_height):
        """Returns the diameter inside bark (DIB) at any given stem height"""
        return math.floor(self.equation(self.dbh, self.height, stem_height, **self.coef[1]))

    def _get_tpa_ba_ac_rd_ac(self):
        """Calculates the Trees per Acre, Basal Area per Acre and Relative Density per Acre
           based on the plot factor"""
        if self.plot_factor == 0:
            return
        elif self.plot_factor > 0:
            self.tpa = self.plot_factor / self.ba
            self.ba_ac = self.plot_factor
            self.rd_ac = self.tpa * self.rd
        else:
            self.tpa = abs(self.plot_factor)
            self.ba_ac = abs(self.plot_factor) * self.ba
            self.rd_ac = self.tpa * self.rd

    def _get_merch_dib(self):
        """Form Percent is the percent of DIB at the Form Height feet above ground,
           this percent will be rounded down for the merch DIB in inches.
           Industry standards are 40% and 17 feet"""
        return math.floor(0.40 * self.get_any_dib(17))

    def _get_merch_height(self):
        """Merch Height is calculated by a Divide and Conquer Algorithm with the starting check height
           at 75% of the total height. The starting merch height is found when the check DIB equals the Merch DIB.
           All DIBs are rounded down to their floor values, so there may be multiple stem heights with the same DIB integer.
           The final merch height will be the top extent of this stem height range"""
        notcheck = True
        floor = 0
        ceiling = self.height
        chkhgt = (ceiling - floor) // 4 * 3

        while notcheck:
            chkdib = int(math.floor(self.equation(self.dbh, self.height, chkhgt, **self.coef[1])))
                
            if chkdib == int(self.merch_dib):
                for i in range(1, 21):
                    chkhgt += 1
                    chkdib_after = int(math.floor(self.equation(self.dbh, self.height, chkhgt, **self.coef[1])))

                    if chkdib_after != chkdib:
                        notcheck = False
                        break
            elif chkdib > int(self.merch_dib):
                floor = chkhgt
                chkhgt = ceiling - ((ceiling - floor) // 2)
            else:
                ceiling = chkhgt
                chkhgt = ceiling - ((ceiling - floor) // 2)
                       
        return chkhgt - 1


    def _get_volume_and_logs(self):
        """Method for cruising the tree, this will determine the stem heights and lengths of the logs, which are sent to
           the Log Class for volume calculations, return a dictionary of the logs by log number"""
        stem_heights = self._calc_stem_heights()
        lengths = [self._calc_log_length(stem_heights[i-1], stem_heights[i]) for i in range(1, len(stem_heights))]
        stem_heights.pop(0)

        logs = {}
        bf = 0
        cf = 0
        for i, (stem_height, length) in enumerate(zip(stem_heights, lengths)):
            log = Log(self, stem_height, length)
            bf += log.bf
            cf += log.cf
            logs[i+1] = log

        self.bf = bf
        self.cf = cf
        self.bf_ac = self.bf * self.tpa
        self.cf_ac = self.cf * self.tpa
        self.vbar = self.bf / self.ba
        return logs

    def _calc_stem_heights(self):
        """Starting at stem height of 1 (stump height), master is updated with the log stem height calculated from
           self._calc_log_stem, if self._calc_log_stem returns None, all logs have been found and iteration is complete"""
        master = [1]
        for i in range(401):
            if not self._calc_log_stem(master[i]):
                break
            else:
                master.append(self._calc_log_stem(master[i]))
        return master

    def _calc_log_stem(self, previous_log_stem_height):
        """Using the previous_log_stem_height arg, it will check if the minimum log length added to previous stem height plus
           1 foot of in-between is greater than the merch height, if it is, it will return None and no more logs can be added. If not
           it will check if the preferred log length added to the previous stem height plus 1 foot of in-between is less than
           or equal to the merch height, if it is then the new stem height is returned and a 40 foot (or user defined preferred length)
           log will added. If not then the merch height is the returned and a final log is added with a length determined by the difference
           between the merch height and previous stem height"""
        min_log_check = previous_log_stem_height + self.min_log + 1

        if min_log_check > self.merch_height - 2:
            return None
        else:
            if previous_log_stem_height + 1 + self.pref_log <= self.merch_height:
                return previous_log_stem_height + self.pref_log + 1
            else:
                return self.merch_height

    def _calc_log_length(self, previous_log_stem_height, current_log_stem_height):
        """Returns a log length in multiples of 2 (24, 26, 28... feet)"""
        return (current_log_stem_height - previous_log_stem_height - 1) // 2 * 2


class TimberFull(object):
    """TimberFull is a class that will cruise a tree based on it's based on the user-cruised logs. These logs can be manually
       added to the class using the add_log method. Required arguments for add_log are stem height (int), log length (int),
       log grade (str), and log defect (int). Log defect should be the whole number value of the estimated percent defect 10% = 10.

       Like TimberQuick, TimberFull uses stem-taper equations from Czaplewski, Kozak, or Wensel
       (depending on species) to calculate the DIB (diameter inside bark) at any stem height.

       TimberFull will instantiate the tree with common individual and per acre metrics based on the input args.

       When the user adds a log using the add_log method, the log metrics are sent to the Log Class,
       to which their volumes in Board Feet (using Scribner Coefficients based on Log Length and top DIB)
       and Cubic Feet (based on the Two-End Conic Cubic Foot Rule) are calculated.

       For inventories, this class is meant to be added to the Plot Class using the Plot Class method of add_tree"""

    def __init__(self, species: str, dbh: float, total_height: int, plot_factor: float):
        self.species = str(species).upper()
        self.dbh = float(dbh)
        self.height = int(total_height)
        self.plot_factor = float(plot_factor)

        self.hdr = self.height / (self.dbh / 12)
        self.ba = self.dbh ** 2 * 0.005454
        self.rd = self.ba / math.sqrt(self.dbh)

        self.coef = ALL_TAPERS_DICT[self.species]
        self.equation = EQUATION_DICT[self.coef[0]]

        self.tpa, self.ba_ac, self.rd_ac = 0, 0, 0
        self._get_tpa_ba_ac_rd_ac()

        self.bf = 0
        self.cf = 0
        self.bf_ac = 0
        self.cf_ac = 0
        self.vbar = 0

        self.logs = {}

    def __getitem__(self, item):
        return self.__dict__[item]

    def add_log(self, stem_height, length, grade, defect):
        """Adds Log Class to the logs dictionary of TimberFull and recalculates the tree's volumes and
           volume-related metrics"""
        if not self.logs:
            self.logs[1] = Log(self, stem_height, length, defect_pct=defect, grade=grade.upper())
        else:
            num = max(self.logs) + 1
            self.logs[num] = Log(self, stem_height, length, defect_pct=defect, grade=grade.upper())
        self._calc_volume_and_logs()

    def get_any_dib(self, stem_height):
        """Returns the diameter inside bark (DIB) at any given stem height"""
        return math.floor(self.equation(self.dbh, self.height, stem_height, **self.coef[1]))

    def _calc_volume_and_logs(self):
        """Calcuates the tree's volume and volume-related metrics based on the log volumes"""
        if not self.logs:
            return
        else:
            bf = 0
            cf = 0
            for lnum in self.logs:
                log = self.logs[lnum]
                bf += log.bf
                cf += log.cf

            self.bf = bf
            self.cf = cf
            self.bf_ac = self.bf * self.tpa
            self.cf_ac = self.cf * self.tpa
            self.vbar = self.bf / self.ba

    def _get_tpa_ba_ac_rd_ac(self):
        """Calculates the Trees per Acre, Basal Area per Acre and Relative Density per Acre
           based on the plot factor"""
        if self.plot_factor == 0:
            return
        elif self.plot_factor > 0:
            self.tpa = self.plot_factor / self.ba
            self.ba_ac = self.plot_factor
            self.rd_ac = self.tpa * self.rd
        else:
            self.tpa = abs(self.plot_factor)
            self.ba_ac = abs(self.plot_factor) * self.ba
            self.rd_ac = self.tpa * self.rd





"""EXAMPLE OF TIMBER CLASSES"""

if __name__ == ('__main__'):

    #tree_data_list -> [Species, DBH, Total Height, Plot Factor]

    tree_data_list = ['PP', 32.2, 145, 33.3]
    tree = TimberQuick(tree_data_list[0], tree_data_list[1], tree_data_list[2], tree_data_list[3])
    print(f'Tree BF: {tree.bf}')
    print(f'Tree RD/ac: {tree.rd_ac}')
    print(f'Tree BF/ac: {tree.bf_ac}')
    print(f'Tree CF/ac: {tree.cf_ac}')
    print('Tree Logs:')
    for lnum in tree.logs:
        print(f'\tLog # {lnum}:')
        log = tree.logs[lnum]
        print(f'\t\tLog Stem Height: {log.stem_height}')
        print(f'\t\tLog Length: {log.length}')
        print(f'\t\tLog Top DIB: {log.top_dib}')
        print(f'\t\tLog Grade: {log.grade}')
        print(f'\t\tLog BF: {log.bf}')
        print(f'\t\tLog CF: {log.cf}\n')

    tree2_data_list = ['DF', 25.5, 125, 40]
    tree2 = TimberFull(tree2_data_list[0], tree2_data_list[1], tree2_data_list[2], tree2_data_list[3])
    tree2.add_log(42, 40, 'SM', 5)
    tree2.add_log(83, 40, 'S2', 0)
    tree2.add_log(102, 18, 'S4', 10)
    print(f'Tree BF: {tree2.bf}')
    print(f'Tree RD/ac: {tree2.rd_ac}')
    print(f'Tree BF/ac: {tree2.bf_ac}')
    print(f'Tree CF/ac: {tree2.cf_ac}')
    print('Tree Logs:')
    for lnum in tree2.logs:
        print(f'\tLog # {lnum}:')
        log = tree2.logs[lnum]
        print(f'\t\tLog Stem Height: {log.stem_height}')
        print(f'\t\tLog Length: {log.length}')
        print(f'\t\tLog Top DIB: {log.top_dib}')
        print(f'\t\tLog Grade: {log.grade}')
        print(f'\t\tLog BF: {log.bf}')
        print(f'\t\tLog CF: {log.cf}\n')





















    



        






    




from fpdf import FPDF
from treetopper._constants import (format_comma,
                                   sort_grade,
                                   SORTED_HEADS,
                                   ALL_SPECIES_NAMES,
                                   LOG_LENGTHS)


class PDF(FPDF):
    def footer(self):
        self.set_y(-15)
        self.set_font('Times', 'I', 8)
        self.cell(0, 10, 'Page ' + str(self.page_no()) + '/{nb}', 0, 0, 'C')

    def compile_report(self, stand):
        self.set_font('Times', 'B', 10)
        height = 8

        self.cell(150, height, f'STAND REPORT FOR: {stand.name}', 0, 0, align='L')
        self.ln(height * 2)

        # Summary Table
        self.cell(50, height, 'STAND CONDITIONS', 0, 0, align='L')

        table_head = ['SPECIES'] + [head[1] for head in SORTED_HEADS]
        self.ln(height)

        spp_col = 18
        end_col = 21
        dat_col = 18
        metric_col = 24

        for i, tab in enumerate(table_head):
            self.set_font('Times', 'B', 7)
            if i == 0:
                self.cell(spp_col, height, tab, 1, 0, align='L')
            elif i >= len(table_head) - 2:
                self.cell(end_col, height, tab, 1, 0, align='C')
            else:
                self.cell(dat_col, height, tab, 1, 0, align='C')

        self.ln(height)

        table_data = self._get_species_table(stand.species)

        for i, row in enumerate(table_data):
            if i == len(table_data) - 1:
                self.set_font('Times', 'B', 8)
            else:
                self.set_font('Times', '', 8)
            for j, col in enumerate(row):
                if j == 0:
                    self.cell(spp_col, height, col, 1, 0, align='L')
                elif j >= len(row) - 2:
                    self.cell(end_col, height, col, 1, 0, align='C')
                else:
                    self.cell(dat_col, height, col, 1, 0, align='C')
            self.ln(height)

        self.add_page()
        self.set_font('Times', 'B', 10)
        self.cell(50, height, 'LOG TABLES', 0, 0, align='L')
        self.ln(height)

        spp_col = metric_col * 7
        log_tables = self._get_log_tables(stand.logs)

        for i, value in enumerate(log_tables):
            for j, table in enumerate(value):
                if j == 0:
                    if i > 0:
                        self.ln(height)
                    self.set_font('Times', 'B', 10)
                    self.cell(spp_col, height, table, 0, 0, align='L')
                    self.ln(height)
                else:
                    for k, row in enumerate(table):
                        if k == 0:
                            self.set_font('Times', 'B', 9)
                            self.cell(spp_col, height, row, 1, 0, align='C')
                            self.ln(height)
                        else:
                            if k == 1:
                                self.set_font('Times', 'B', 7)
                            elif k == len(table) - 1:
                                self.set_font('Times', 'B', 8)
                            else:
                                self.set_font('Times', '', 8)

                            for l, col in enumerate(row):
                                if l == 0:
                                    self.cell(metric_col, height, col, 1, 0, align='L')
                                else:
                                    self.cell(metric_col, height, col, 1, 0, align='C')
                            self.ln(height)
                    self.ln(height)
            self.add_page()
            if i < len(log_tables) - 1:
                self.set_font('Times', 'B', 10)
                self.cell(50, height, 'LOG TABLES', 0, 0, align='L')

        self.set_font('Times', 'B', 10)
        self.cell(50, height, 'STAND STATISTICS', 0, 0, align='L')
        self.ln(height)

        metric_col = 24
        spp_width = metric_col + ((end_col - 1) * 8)
        stats_data = self._get_species_stats_table(stand.species_stats)
        for i, table in enumerate(stats_data):
            for j, row in enumerate(table):
                if j == 0:
                    self.set_font('Times', 'B', 9)
                    self.cell(spp_width, height, row, 1, 0, align='C')
                else:
                    if j == 1:
                        self.set_font('Times', 'B', 7)
                    else:
                        self.set_font('Times', '', 8)
                    for k, col in enumerate(row):
                        if k == 0:
                            self.cell(metric_col, height, col, 1, 0, align='L')
                        else:
                            self.cell(end_col-1, height, col, 1, 0, align='C')
                self.ln(height)
            self.ln(height)

    def _get_species_table(self, species):
            table_data = []
            for spp in species:
                temp = []
                if spp == 'totals_all':
                    temp.append('TOTALS')
                else:
                    temp.append(spp)
                for i, sub in enumerate(SORTED_HEADS):
                    temp.append(format_comma(species[spp][sub[0]]))
                table_data.append(temp)
            table_data.append(table_data.pop(0))
            return table_data

    def _get_log_tables(self, logs):
        heads = ['LOG LENGTHS'] + [rng.upper() for rng in LOG_LENGTHS] + ['TOTALS']
        tables = [['bf_ac', 'BOARD FEET PER ACRE'], ['cf_ac', 'CUBIC FEET PER ACRE'], ['lpa', 'LOGS PER ACRE']]
        table_data = []

        for i in tables:
            temp = [i[1]]
            for species in logs:
                temp_temp = []
                if species == 'totals_all':
                    label = 'TOTAL SPECIES'
                else:
                    label = ALL_SPECIES_NAMES[species]

                temp_temp.append(label)
                temp_temp.append(heads)
                grade_sort = []
                for grade in logs[species]:
                    values = [logs[species][grade][rng][i[0]]['mean'] for rng in logs[species][grade]]
                    if sum(values) > 0:
                        if grade == 'totals_by_length':
                            txt = 'TOTALS'
                        else:
                            txt = grade
                        show2 = [txt] + [format_comma(z) for z in values]
                        grade_sort.append(show2)
                grade_sort = sorted(grade_sort, key=sort_grade)
                for g in grade_sort:
                    temp_temp.append(g)
                temp.append(temp_temp)

            if len(logs) <= 2:
                temp.pop(1)
            else:
                temp.append(temp.pop(1))

            table_data.append(temp)

        return table_data

    def _get_species_stats_table(self, species_stats):
        labels = {'tpa': 'TPA',
                  'ba_ac': 'BASAL AREA',
                  'rd_ac': 'RD',
                  'bf_ac': 'BOARD FEET',
                  'cf_ac': 'CUBIC FEET'}
        table_data = []
        for spp in species_stats:
            if spp == 'totals_all':
                show = 'TOTALS'
            else:
                show = ALL_SPECIES_NAMES[spp]
            temp = [show]
            heads = ['METRIC'] + [head.upper() for head in species_stats[spp]['tpa'] if head != 'low_avg_high'] + ['LOW', 'AVERAGE', 'HIGH']
            temp.append(heads)
            for key in species_stats[spp]:
                temp_temp = [labels[key]]
                for sub in species_stats[spp][key]:
                    data = species_stats[spp][key][sub]
                    if data == 'Not enough data':
                        temp_temp.append(data)
                        for i in range(6):
                            temp_temp.append('-')
                        break
                    else:
                        if sub != 'stderr_pct' and sub != 'low_avg_high':
                            temp_temp.append(format_comma(data))
                        elif sub == 'stderr_pct':
                            temp_temp.append(str(round(data, 1)) + ' %')
                        else:
                            for i in data:
                                temp_temp.append(format_comma(i))
                temp.append(temp_temp)
            table_data.append(temp)

        if len(species_stats) <= 2:
            table_data.pop(0)
        else:
            table_data.append(table_data.pop(0))

        return table_data

    def compile_thin_report(self, thin, message=None):
        self.set_font('Times', 'B', 10)
        height = 8
        species, min_dbh, max_dbh = thin._get_message_params_report()
        self.cell(150, height, f'THINNING TARGET: {thin.target} {thin.target_metric.replace("_", "/").upper()}', 0, 0, align='L')
        self.ln(height)
        self.cell(150, height, f'THINNING PARAMETERS:', 0, 0, align='L')
        self.ln(height)

        self.set_font('Times', 'B', 8)
        self.cell(150, height, f'    SPECIES: {", ".join(species)}', 0, 0, align='L')
        self.ln(height)
        self.cell(150, height, f'    MINIMUM DBH: {min_dbh}', 0, 0, align='L')
        self.ln(height)
        self.cell(150, height, f'    MAXIMUM DBH: {max_dbh}', 0, 0, align='L')
        self.ln(height * 2)

        spp_col = 18
        end_col = 21
        dat_col = 18
        metric_col = 24

        table_data = self._get_thin_species_table(thin.species_data)

        for i, condition in enumerate(table_data):
            for j, row in enumerate(condition):
                if j == 0:
                    self.set_font('Times', 'B', 10)
                    self.cell(50, height, f'{row}', 0, 0, align='L')
                    self.ln(height)

                    table_head = ['SPECIES'] + [head[1] for head in SORTED_HEADS]
                    for i, tab in enumerate(table_head):
                        self.set_font('Times', 'B', 7)
                        if i == 0:
                            self.cell(spp_col, height, tab, 1, 0, align='L')
                        elif i >= len(table_head) - 2:
                            self.cell(end_col, height, tab, 1, 0, align='C')
                        else:
                            self.cell(dat_col, height, tab, 1, 0, align='C')
                    self.ln(height)
                else:
                    if j == len(condition) - 1:
                        self.set_font('Times', 'B', 8)
                    else:
                        self.set_font('Times', '', 8)
                    for k, col in enumerate(row):
                        if k == 0:
                            self.cell(spp_col, height, col, 1, 0, align='L')
                        elif k >= len(row) - 2:
                            self.cell(end_col, height, col, 1, 0, align='C')
                        else:
                            self.cell(dat_col, height, col, 1, 0, align='C')
                    self.ln(height)
            self.ln(height)

        if message:
            self.set_font('Times', '', 10)
            for line in message:
                self.cell(250, height, line, 0, 0, align='L')
                self.ln(height)



    def _get_thin_species_table(self, species):
        table_data = []
        for condition in species:
            temp = [condition.replace('_', '').upper()]
            for spp in species[condition]:
                temp_temp = []
                if spp == 'totals_all':
                    temp_temp.append('TOTALS')
                else:
                    temp_temp.append(spp)
                for i, sub in enumerate(SORTED_HEADS):
                    temp_temp.append(format_comma(species[condition][spp][sub[0]]))
                temp.append(temp_temp)
            temp.append(temp.pop(1))
            table_data.append(temp)
        table_data.append(table_data.pop(1))
        return table_data



import unittest

import win32com.client
import win32com.test.util
import win32com.server.util

class Tester:
    _public_methods_ = [ 'TestValue' ]
    def TestValue(self, v):
        pass


def test_ob():
    return win32com.client.Dispatch(win32com.server.util.wrap(Tester()))

class TestException(Exception):
    pass

# The object we try and pass - pywin32 will call __float__ as a last resort.
class BadConversions:
    def __float__(self):
        raise TestException()

class TestCase(win32com.test.util.TestCase):
    def test_float(self):
        try:
            test_ob().TestValue(BadConversions())
            raise Exception("Should not have worked")
        except Exception as e:
            assert isinstance(e, TestException)

if __name__=='__main__':
    unittest.main()



from treetopper._constants import (math,
                                   GRADE_NAMES,
                                   LOG_LENGTHS,
                                   OFFICIAL_GRADES,
                                   SCRIBNER_DICT)



class Log(object):
    """Log Class calculates the volume of an individual log from the Timber Classes. If coming from
       the TimberQuick class, it will also calculate the grade of the log"""

    def __init__(self, timber, stem_height: int, length: int, defect_pct: int = 0, grade: str = None):
        self.tree = timber
        self.stem_height = stem_height
        self.length = length
        self.defect = defect_pct

        self.species = self.tree.species
        self.lpa = self.tree.tpa

        self.top_dib = self._calc_top_dib()
        if grade:
            self.grade = grade
        else:
            self.grade = self._calc_log_grade()
        self.scrib = self._calc_scribner()
        self.bf = self._calc_board_feet()
        self.cf = self._calc_cubic_feet()

        self.bf_ac = self.bf * self.lpa
        self.cf_ac = self.cf * self.lpa


        self.grade_name = GRADE_NAMES[self.grade]
        self.length_range = self._get_length_range()


    def __getitem__(self, attribute: str):
        return self.__dict__[attribute]

    def _get_length_range(self):
        """Gets the length range that the logs length is in, this is used for the Stand Class' report methods"""
        for rng in LOG_LENGTHS:
            if LOG_LENGTHS[rng][0] <= self.length <= LOG_LENGTHS[rng][1]:
                return rng

    def _calc_top_dib(self):
        """Uses the Timber Class' stem taper equation to calculate the top diameter inside bark (DIB) of the log
           from the stem_height argument"""
        return int(math.floor(self.tree.equation(self.tree.dbh, self.tree.height, self.stem_height, **self.tree.coef[1])))

    def _calc_log_grade(self):
        """Used when the Timber Class is TimberQuick, it will get the grade of the log based on species, minimum log lengths,
           and minimum log top DIBs set forth in the Official Rules for the Log Scaling and Grading Bureaus"""
        for i in OFFICIAL_GRADES[self.species]:
            if self.top_dib >= i[0] and self.length >= i[1]:
                return i[2]

    def _calc_board_feet(self):
        """Returns the board feet of the log based on log length and the corresponding Scribner coefficient"""
        return math.floor(self.length * self.scrib * (1 - (self.defect / 100)))

    def _calc_cubic_feet(self):
        """Return the cubic feet of the log based on the Two-End Conic Cubic Foot Rule"""
        if self.length < 17:
            x = self.length * 0.67
        else:
            x = self.length + 1
        return ((.005454 * x) * (((2 * ((self.top_dib + 0.7) ** 2)) + (2 * (self.top_dib + 0.7))) / 3)) * (1 - (self.defect / 100))

    def _calc_scribner(self):
        """Return the Scribner coefficient for board foot calculation based on log length and log top DIB"""
        if self.top_dib in range(6, 12):
            if 0 < self.length < 16:
                return SCRIBNER_DICT[self.top_dib][0]
            elif 16 <= self.length < 32:
                return SCRIBNER_DICT[self.top_dib][1]
            else:
                return SCRIBNER_DICT[self.top_dib][2]
        else:
            return SCRIBNER_DICT[self.top_dib]





from treetopper._constants import (format_comma,
                                   sort_grade,
                                   LOG_LENGTHS,
                                   ALL_SPECIES_NAMES,
                                   SORTED_HEADS)

SPACE = 20


def print_species(species):
    heads = ['SPECIES'] + [head[1] for head in SORTED_HEADS]
    formatted_heads = ''.join([head + (' ' * (SPACE - len(head))) for head in heads])
    formatted_data = []
    for z, key in enumerate(species):
        if key == 'totals_all':
            show = 'TOTALS'
        else:
            show = key
        temp = [str(show)] + [format_comma(species[key][i[0]]) for i in SORTED_HEADS]
        formatted_data.append(''.join([t + (' ' * (SPACE - len(t))) for t in temp]))
        if z == len(species) - 1:
            formatted_data.append('-' * (SPACE * len(heads)))

    formatted_data.append(formatted_data.pop(0))

    text = f'{formatted_heads}\n'
    text += f'{"-" * SPACE * len(heads)}\n'
    for i in formatted_data:
        text += f'{i}\n'
    return text


def print_logs(logs):
    heads = ['LOG LENGTHS'] + [rng.upper() for rng in LOG_LENGTHS] + ['TOTALS']
    formatted_heads = ''.join([head + (' ' * (SPACE - len(head))) for head in heads])
    tables = [['bf_ac', 'BOARD FEET PER ACRE'], ['cf_ac', 'CUBIC FEET PER ACRE'], ['lpa', 'LOGS PER ACRE']]
    table_data = []

    for i in tables:
        temp = [[i[1]]]
        for species in logs:
            temp_temp = []
            ratio = 56
            if species == 'totals_all':
                label = 'TOTAL SPECIES'
            else:
                label = ALL_SPECIES_NAMES[species]
            first = '-' * (((SPACE * len(heads)) - len(label)) // 100 * ratio)
            second = '-' * ((len(heads) * SPACE) - (len(first) + len(label)))
            show = [first + label + second]

            temp_temp.append(''.join(show))
            temp_temp.append(formatted_heads)
            temp_temp.append('-' * (SPACE * len(heads)))
            grade_sort = []
            for grade in logs[species]:
                values = [logs[species][grade][rng][i[0]]['mean'] for rng in logs[species][grade]]
                if sum(values) > 0:
                    if grade == 'totals_by_length':
                        txt = 'TOTALS'
                    else:
                        txt = grade
                    show2 = [txt] + [format_comma(z) for z in values]
                    grade_sort.append(show2)
            grade_sort = sorted(grade_sort, key=sort_grade)
            for g in grade_sort:
                temp_temp.append(''.join([gs + (' ' * (SPACE - len(gs))) for gs in g]))
            temp.append(temp_temp)

        if len(logs) <= 2:
            temp.pop(1)
        else:
            temp.append(temp.pop(1))

        table_data.append(temp)

    text = ''
    for i in table_data:
        for z, j in enumerate(i):
            for k in j:
                text += f'{k}\n'
            if z > 0:
                if z == len(i) - 1:
                    text += f'{"-" * SPACE * len(heads)}\n'
                else:
                    text += '\n'
        text += '\n\n'
    return text


def print_species_stats(species):
    print_data = []
    for spp in species:
        if spp == 'totals_all':
            show = 'TOTALS'
        else:
            show = ALL_SPECIES_NAMES[spp]
        temp = [show]
        heads = ['METRIC'] + [head.upper() for head in species[spp]['tpa'] if head != 'low_avg_high'] + ['LOW', 'AVERAGE', 'HIGH']
        formatted_heads = ''.join([head + (' ' * (SPACE - len(head))) for head in heads])
        temp.append(formatted_heads)
        temp.append('-' * (SPACE * len(heads)))

        for key in species[spp]:
            temp_temp = [key.upper() + (' ' * (SPACE - len(key)))]
            for sub in species[spp][key]:
                data = species[spp][key][sub]
                if data == 'Not enough data':
                    temp_temp.append(data)
                    for i in range(6):
                        temp_temp.append('-')
                    break
                else:
                    if sub != 'stderr_pct' and sub != 'low_avg_high':
                        temp_temp.append(format_comma(data))
                    elif sub == 'stderr_pct':
                        temp_temp.append(str(round(data, 1)) + ' %')
                    else:
                        for i in data:
                            temp_temp.append(format_comma(i))
            temp.append(''.join([dat + (' ' * (SPACE - len(dat))) for dat in temp_temp]))
        print_data.append(temp)

    if len(species) <= 2:
        print_data.pop(0)
    else:
        print_data.append(print_data.pop(0))

    text = ''
    for i in print_data:
        for j in i:
            text += f'\n{j}'
        text += '\n'
    return text


def print_thin_species(species):
    all_data = []
    for condition in species:
        formatted_data = [condition.replace('_', '').upper()]
        heads = ['SPECIES'] + [head[1] for head in SORTED_HEADS]
        formatted_heads = ''.join([head + (' ' * (SPACE - len(head))) for head in heads])
        formatted_data.append(formatted_heads)
        formatted_data.append('-' * (SPACE * len(heads)))
        for z, spp in enumerate(species[condition]):
            if spp == 'totals_all':
                show = 'TOTALS'
            else:
                show = spp
            temp = [str(show)] + [format_comma(species[condition][spp][i[0]]) for i in SORTED_HEADS]
            formatted_data.append(''.join([t + (' ' * (SPACE - len(t))) for t in temp]))
            if z == len(species[condition]) - 1:
                formatted_data.append('-' * (SPACE * len(heads)))

        formatted_data.append(formatted_data.pop(3))
        all_data.append(formatted_data)

    all_data.append(all_data.pop(1))

    text = ''

    for i in all_data:
        for j in i:
            #print(j)
            text += f'\n{j}'
        text += '\n\n'
        #print('\n')
    return text


def print_plot_logs(plot):
    logs = plot.logs
    heads = ['LOG LENGTHS'] + [rng.upper() for rng in LOG_LENGTHS] + ['TOTALS']
    formatted_heads = ''.join([head + (' ' * (SPACE - len(head))) for head in heads])
    tables = [['bf_ac', 'BOARD FEET PER ACRE']]#, ['cf_ac', 'CUBIC FEET PER ACRE'], ['lpa', 'LOGS PER ACRE']]
    table_data = []

    for i in tables:
        temp = [[i[1]]]
        for species in logs:
            temp_temp = []
            ratio = 56
            if species == 'totals_all':
                label = 'TOTAL SPECIES'
            else:
                label = ALL_SPECIES_NAMES[species]
            first = '-' * (((SPACE * len(heads)) - len(label)) // 100 * ratio)
            second = '-' * ((len(heads) * SPACE) - (len(first) + len(label)))
            show = [first + label + second]

            temp_temp.append(''.join(show))
            temp_temp.append(formatted_heads)
            temp_temp.append('-' * (SPACE * len(heads)))
            grade_sort = []
            for grade in logs[species]:
                if grade == 'totals_by_length':
                    txt = 'TOTALS'
                else:
                    txt = grade
                show2 = [txt] + [format_comma(logs[species][grade][rng][i[0]]) for rng in logs[species][grade] if rng != 'display']
                grade_sort.append(show2)
            grade_sort = sorted(grade_sort, key=sort_grade)
            for g in grade_sort:
                temp_temp.append(''.join([gs + (' ' * (SPACE - len(gs))) for gs in g]))
            temp.append(temp_temp)

        if len(logs) <= 2:
            temp.pop(1)
        else:
            temp.append(temp.pop(1))

        table_data.append(temp)

    for i in table_data:
        for z, j in enumerate(i):
            for k in j:
                print(k)
            if z > 2:
                print()
        print('\n')





if __name__ == '__main__':
    x = 10
    for i in range(15):
        print(format_comma(x + .1))
        x = x * 10




# Configure this to _YOUR_ environment in order to run the testcases.
"testADOdbapiConfig.py v 2.6.2.B00"

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# #
# #  TESTERS:
# #
# #  You will need to make numerous modifications to this file
# #  to adapt it to your own testing environment.
# #
# #  Skip down to the next "# #" line --
# #  -- the things you need to change are below it.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
import platform
import sys
import random

import is64bit
import setuptestframework
import tryconnection

print("\nPython", sys.version)
node = platform.node()
try: print('node=%s, is64bit.os()= %s, is64bit.Python()= %s' % (node, is64bit.os(), is64bit.Python()))
except: pass

if '--help' in sys.argv:
    print("""Valid command-line switches are:
    --package - create a temporary test package, run 2to3 if needed.
    --all - run all possible tests
    --time - loop over time format tests (including mxdatetime if present)
    --nojet - do not test against an ACCESS database file
    --mssql - test against Microsoft SQL server
    --pg - test against PostgreSQL
    --mysql - test against MariaDB
    --remote= - test unsing remote server at= (experimental) 
    """)
    exit()
try:
    onWindows = bool(sys.getwindowsversion()) # seems to work on all versions of Python
except:
    onWindows = False

# create a random name for temporary table names
_alphabet = "PYFGCRLAOEUIDHTNSQJKXBMWVZ" # why, yes, I do happen to use a dvorak keyboard
tmp = ''.join([random.choice(_alphabet) for x in range(9)])
mdb_name = 'xx_' + tmp + '.mdb'  # generate a non-colliding name for the temporary .mdb
testfolder = setuptestframework.maketemp()

if '--package' in sys.argv:
    #  create a new adodbapi module -- running 2to3 if needed.
    pth = setuptestframework.makeadopackage(testfolder)
else:
    #  use the adodbapi module in which this file appears
    pth = setuptestframework.find_ado_path()
if pth not in sys.path:
    #  look here _first_ to find modules
    sys.path.insert(1,pth)

proxy_host = None
for arg in sys.argv:
    if arg.startswith('--remote='):
        proxy_host = arg.split("=")[1]
        import adodbapi.remote as remote
        break


# function to clean up the temporary folder -- calling program must run this function before exit.
cleanup = setuptestframework.getcleanupfunction()
try:
    import adodbapi  # will (hopefully) be imported using the "pth" discovered above
except SyntaxError:
    print('\n* * * Are you trying to run Python2 code using Python3? Re-run this test using the "--package" switch.')
    sys.exit(11)
try:
    print(adodbapi.version) # show version
except:
    print('"adodbapi.version" not present or not working.')
print(__doc__)

verbose = False
for a in sys.argv:
    if a.startswith('--verbose'):
        arg = True
        try: arg = int(a.split("=")[1])
        except IndexError: pass
        adodbapi.adodbapi.verbose = arg
        verbose = arg

doAllTests = '--all' in sys.argv
doAccessTest = not ('--nojet' in sys.argv)
doSqlServerTest = '--mssql' in  sys.argv or doAllTests
doMySqlTest = '--mysql' in sys.argv or doAllTests
doPostgresTest = '--pg' in sys.argv or doAllTests
iterateOverTimeTests = ('--time' in sys.argv or doAllTests) and onWindows

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# # start your environment setup here v v v
SQL_HOST_NODE = 'testsql.2txt.us,1430'

try: #If mx extensions are installed, use mxDateTime
    import mx.DateTime
    doMxDateTimeTest=True
except: 
    doMxDateTimeTest=False #Requires eGenixMXExtensions

doTimeTest = True # obsolete python time format

if doAccessTest:
    if proxy_host:  # determine the (probably remote) database file folder
        c = {'macro_find_temp_test_path': ['mdb', mdb_name],
             'proxy_host': proxy_host }
    else:
        c = {'mdb': setuptestframework.makemdb(testfolder, mdb_name)}

    # macro definition for keyword "provider"  using macro "is64bit" -- see documentation
    # is64bit will return true for 64 bit versions of Python, so the macro will select the ACE provider
    # (If running a remote ADO service, this will test the 64-bitedness of the ADO server.)
    c['macro_is64bit'] = ['provider',
                          "Microsoft.ACE.OLEDB.12.0",  # 64 bit provider
                          "Microsoft.Jet.OLEDB.4.0"]   # 32 bit provider
    connStrAccess = "Provider=%(provider)s;Data Source=%(mdb)s"  # ;Mode=ReadWrite;Persist Security Info=False;Jet OLEDB:Bypass UserInfo Validation=True"
    print('    ...Testing ACCESS connection to {} file...'.format(c.get('mdb', 'remote .mdb')))
    doAccessTest, connStrAccess, dbAccessconnect = tryconnection.try_connection(verbose, connStrAccess, 10, **c)

if doSqlServerTest:
    c = {'host': SQL_HOST_NODE,  # name of computer with SQL Server
        'database': "adotest",
        'user' : 'adotestuser',   # None implies Windows security
        'password' : "Sq1234567",
        # macro definition for keyword "security" using macro "auto_security"
        'macro_auto_security' : 'security',
        'provider' : 'MSOLEDBSQL; MARS Connection=True'
         }
    if proxy_host:
        c['proxy_host'] = proxy_host
    connStr = "Provider=%(provider)s; Initial Catalog=%(database)s; Data Source=%(host)s; %(security)s;"
    print('    ...Testing MS-SQL login to {}...'.format(c['host']))
    doSqlServerTest, connStrSQLServer, dbSqlServerconnect = tryconnection.try_connection(verbose, connStr, 30, **c)

if doMySqlTest:
    c = {'host' : "testmysql.2txt.us",
        'database' : 'adodbapitest',
        'user' : 'adotest',
        'password' : '12345678',
         'port'    : '3330',  # note the nonstandard port for obfuscation
        'driver' : "MySQL ODBC 5.1 Driver"}    # or _driver="MySQL ODBC 3.51 Driver
    if proxy_host:
        c['proxy_host'] = proxy_host
    c['macro_is64bit'] = ['provider', 'Provider=MSDASQL;']  # turn on the 64 bit ODBC adapter only if needed
    cs = '%(provider)sDriver={%(driver)s};Server=%(host)s;Port=3330;' + \
        'Database=%(database)s;user=%(user)s;password=%(password)s;Option=3;'
    print('    ...Testing MySql login to {}...'.format(c['host']))
    doMySqlTest, connStrMySql, dbMySqlconnect = tryconnection.try_connection(verbose, cs, 5, **c)



if doPostgresTest:
    _computername = "testpg.2txt.us"
    _databasename='adotest'
    _username = 'adotestuser'
    _password = '12345678'
    kws = {'timeout': 4}
    kws['macro_is64bit'] = ['prov_drv',
                            'Provider=MSDASQL;Driver={PostgreSQL Unicode(x64)}',
                            'Driver=PostgreSQL Unicode']
    # get driver from http://www.postgresql.org/ftp/odbc/versions/
    # test using positional and keyword arguments (bad example for real code)
    if proxy_host:
        kws['proxy_host'] = proxy_host
    print('    ...Testing PostgreSQL login to {}...'.format(_computername))
    doPostgresTest, connStrPostgres, dbPostgresConnect = tryconnection.try_connection(verbose,
        '%(prov_drv)s;Server=%(host)s;Database=%(database)s;uid=%(user)s;pwd=%(password)s;port=5430;',  # note nonstandard port
         _username, _password, _computername, _databasename, **kws)

assert doAccessTest or doSqlServerTest or doMySqlTest or doPostgresTest, 'No database engine found for testing'



"""Test win32profile"""
import os
import unittest
import win32profile

class Tester(unittest.TestCase):
    def test_environment(self):
        os.environ["FOO"] = "bar=baz"
        env = win32profile.GetEnvironmentStrings()
        assert "FOO" in env
        assert env["FOO"] == "bar=baz"
        assert os.environ["FOO"] == "bar=baz"

if __name__ == '__main__':
    unittest.main()



"""adodbapi - A python DB API 2.0 (PEP 249) interface to Microsoft ADO

Copyright (C) 2002 Henrik Ekelund, version 2.1 by Vernon Cole
* http://sourceforge.net/projects/adodbapi
"""
import sys
import time

from .apibase import apilevel, threadsafety, paramstyle
from .apibase import Warning, Error, InterfaceError, DatabaseError, DataError, OperationalError, IntegrityError
from .apibase import InternalError, ProgrammingError, NotSupportedError, FetchFailedError
from .apibase import NUMBER, STRING, BINARY, DATETIME, ROWID

from .adodbapi import connect, Connection, __version__, dateconverter, Cursor

def Binary(aString):
    """This function constructs an object capable of holding a binary (long) string value. """
    return bytes(aString)

def Date(year,month,day):
    "This function constructs an object holding a date value. "
    return dateconverter.Date(year,month,day)

def Time(hour,minute,second):
    "This function constructs an object holding a time value. "
    return dateconverter.Time(hour,minute,second)

def Timestamp(year,month,day,hour,minute,second):
    "This function constructs an object holding a time stamp value. "
    return dateconverter.Timestamp(year,month,day,hour,minute,second)

def DateFromTicks(ticks):
    """This function constructs an object holding a date value from the given ticks value
    (number of seconds since the epoch; see the documentation of the standard Python time module for details). """
    return Date(*time.gmtime(ticks)[:3])

def TimeFromTicks(ticks):
    """This function constructs an object holding a time value from the given ticks value
    (number of seconds since the epoch; see the documentation of the standard Python time module for details). """
    return Time(*time.gmtime(ticks)[3:6])

def TimestampFromTicks(ticks):
    """This function constructs an object holding a time stamp value from the given
    ticks value (number of seconds since the epoch;
    see the documentation of the standard Python time module for details). """
    return Timestamp(*time.gmtime(ticks)[:6])

version = 'adodbapi v' + __version__



import adodbapi
import datetime
try:
    import adodbapi.is64bit as is64bit
    is64 = is64bit.Python()
except ImportError:
    is64 = False  # in case the user has an old version of adodbapi
if is64:
    driver = "Microsoft.ACE.OLEDB.12.0"
else:
    driver = "Microsoft.Jet.OLEDB.4.0"
filename = 'xx.xls'  # file will be created if it does not exist
extended = 'Extended Properties="Excel 8.0;Readonly=False;"'

constr = "Provider=%s;Data Source=%s;%s" % (driver, filename, extended)

conn = adodbapi.connect(constr)
with conn: # will auto commit if no errors
    with conn.cursor() as crsr:
        try:    crsr.execute('drop table SheetOne')
        except: pass  # just is case there is one already there

        # create the sheet and the header row and set the types for the columns
        crsr.execute('create table SheetOne (Name varchar, Rank varchar, SrvcNum integer, Weight float,  Birth date)')

        sql = "INSERT INTO SheetOne (name, rank , srvcnum, weight, birth) values (?,?,?,?,?)"

        data = ('Mike Murphy', 'SSG', 123456789, 167.8, datetime.date(1922,12,27))
        crsr.execute(sql, data)  # write the first row of data
        crsr.execute(sql, ['John Jones', 'Pvt', 987654321, 140.0, datetime.date(1921,7,4)])  # another row of data
conn.close()
print('Created spreadsheet=%s worksheet=%s' % (filename, 'SheetOne'))



"""adodbapi.apibase - A python DB API 2.0 (PEP 249) interface to Microsoft ADO

Copyright (C) 2002 Henrik Ekelund, version 2.1 by Vernon Cole
* http://sourceforge.net/projects/pywin32
* http://sourceforge.net/projects/adodbapi
"""

import sys
import time
import datetime
import decimal
import numbers
# noinspection PyUnresolvedReferences
from . import ado_consts as adc

verbose = False  # debugging flag

onIronPython = sys.platform == 'cli'
if onIronPython:  # we need type definitions for odd data we may need to convert
    # noinspection PyUnresolvedReferences
    from System import DBNull, DateTime
    NullTypes = (type(None), DBNull)
else:
    DateTime = type(NotImplemented)  # should never be seen on win32
    NullTypes = type(None)

# --- define objects to smooth out Python3 <-> Python 2.x differences
unicodeType = str
longType = int
StringTypes = str
makeByteBuffer = bytes
memoryViewType = memoryview
_BaseException = Exception

try:                #jdhardy -- handle bytes under IronPython & Py3
    bytes
except NameError:
    bytes = str     # define it for old Pythons

# ------- Error handlers ------
def standardErrorHandler(connection, cursor, errorclass, errorvalue):
    err = (errorclass, errorvalue)
    try:
        connection.messages.append(err)
    except: pass
    if cursor is not None:
        try:
            cursor.messages.append(err)
        except: pass
    raise errorclass(errorvalue)

# Note: _BaseException is defined differently between Python 2.x and 3.x
class Error(_BaseException):
    pass   #Exception that is the base class of all other error
    #exceptions. You can use this to catch all errors with one
    #single 'except' statement. Warnings are not considered
    #errors and thus should not use this class as base. It must
    #be a subclass of the Python StandardError (defined in the
    #module exceptions).
class Warning(_BaseException):
    pass
class InterfaceError(Error):
    pass
class DatabaseError(Error):
    pass
class InternalError(DatabaseError):
    pass
class OperationalError(DatabaseError):
    pass
class ProgrammingError(DatabaseError):
    pass
class IntegrityError(DatabaseError):
    pass
class DataError(DatabaseError):
    pass
class NotSupportedError(DatabaseError):
    pass

class FetchFailedError(OperationalError):
    """
    Error is used by RawStoredProcedureQuerySet to determine when a fetch
    failed due to a connection being closed or there is no record set
    returned. (Non-standard, added especially for django)
    """
    pass
# # # # # ----- Type Objects and Constructors ----- # # # # #
#Many databases need to have the input in a particular format for binding to an operation's input parameters.
#For example, if an input is destined for a DATE column, then it must be bound to the database in a particular
#string format. Similar problems exist for "Row ID" columns or large binary items (e.g. blobs or RAW columns).
#This presents problems for Python since the parameters to the executeXXX() method are untyped.
#When the database module sees a Python string object, it doesn't know if it should be bound as a simple CHAR
#column, as a raw BINARY item, or as a DATE.
#
#To overcome this problem, a module must provide the constructors defined below to create objects that can
#hold special values. When passed to the cursor methods, the module can then detect the proper type of
#the input parameter and bind it accordingly.

#A Cursor Object's description attribute returns information about each of the result columns of a query.
#The type_code must compare equal to one of Type Objects defined below. Type Objects may be equal to more than
#one type code (e.g. DATETIME could be equal to the type codes for date, time and timestamp columns;
#see the Implementation Hints below for details).

#SQL NULL values are represented by the Python None singleton on input and output.

#Note: Usage of Unix ticks for database interfacing can cause troubles because of the limited date range they cover.

# def Date(year,month,day):
#     "This function constructs an object holding a date value. "
#     return dateconverter.date(year,month,day)  #dateconverter.Date(year,month,day)
#
# def Time(hour,minute,second):
#     "This function constructs an object holding a time value. "
#     return dateconverter.time(hour, minute, second) # dateconverter.Time(hour,minute,second)
#
# def Timestamp(year,month,day,hour,minute,second):
#     "This function constructs an object holding a time stamp value. "
#     return dateconverter.datetime(year,month,day,hour,minute,second)
#
# def DateFromTicks(ticks):
#     """This function constructs an object holding a date value from the given ticks value
#     (number of seconds since the epoch; see the documentation of the standard Python time module for details). """
#     return Date(*time.gmtime(ticks)[:3])
#
# def TimeFromTicks(ticks):
#     """This function constructs an object holding a time value from the given ticks value
#     (number of seconds since the epoch; see the documentation of the standard Python time module for details). """
#     return Time(*time.gmtime(ticks)[3:6])
#
# def TimestampFromTicks(ticks):
#     """This function constructs an object holding a time stamp value from the given
#     ticks value (number of seconds since the epoch;
#     see the documentation of the standard Python time module for details). """
#     return Timestamp(*time.gmtime(ticks)[:6])
#
# def Binary(aString):
#     """This function constructs an object capable of holding a binary (long) string value. """
#     b = makeByteBuffer(aString)
#     return b
# -----     Time converters ----------------------------------------------
class TimeConverter(object):  # this is a generic time converter skeleton
    def __init__(self):       # the details will be filled in by instances
        self._ordinal_1899_12_31=datetime.date(1899,12,31).toordinal()-1
                    # Use cls.types to compare if an input parameter is a datetime
        self.types = {type(self.Date(2000,1,1)),
                      type(self.Time(12,1,1)),
                      type(self.Timestamp(2000,1,1,12,1,1)),
                      datetime.datetime,
                      datetime.time,
                      datetime.date}
    def COMDate(self,obj):
        '''Returns a ComDate from a date-time'''
        try: # most likely a datetime
            tt=obj.timetuple()

            try:
                ms=obj.microsecond
            except:
                ms=0
            return self.ComDateFromTuple(tt, ms)
        except:  # might be a tuple
            try:
                return self.ComDateFromTuple(obj)
            except: # try an mxdate
                try:
                    return obj.COMDate()
                except:
                    raise ValueError('Cannot convert "%s" to COMdate.' % repr(obj))
    def ComDateFromTuple(self, t, microseconds=0):
        d = datetime.date(t[0],t[1],t[2])
        integerPart = d.toordinal() - self._ordinal_1899_12_31
        ms = (t[3]*3600 + t[4]*60 + t[5]) * 1000000 + microseconds
        fractPart = float(ms) / 86400000000.0
        return integerPart + fractPart
    def DateObjectFromCOMDate(self,comDate):
        'Returns an object of the wanted type from a ComDate'
        raise NotImplementedError   #"Abstract class"
    def Date(self,year,month,day):
        "This function constructs an object holding a date value. "
        raise NotImplementedError   #"Abstract class"
    def Time(self,hour,minute,second):
        "This function constructs an object holding a time value. "
        raise NotImplementedError   #"Abstract class"
    def Timestamp(self,year,month,day,hour,minute,second):
        "This function constructs an object holding a time stamp value. "
        raise NotImplementedError   #"Abstract class"
        # all purpose date to ISO format converter
    def DateObjectToIsoFormatString(self, obj):
        "This function should return a string in the format 'YYYY-MM-dd HH:MM:SS:ms' (ms optional) "
        try:  # most likely, a datetime.datetime
            s = obj.isoformat(' ')
        except (TypeError, AttributeError):
            if isinstance(obj, datetime.date):
                s = obj.isoformat() + ' 00:00:00'  # return exact midnight
            else:
                try: # maybe it has a strftime method, like mx
                    s = obj.strftime('%Y-%m-%d %H:%M:%S')
                except AttributeError: 
                    try:  #but may be time.struct_time
                        s = time.strftime('%Y-%m-%d %H:%M:%S', obj)
                    except:
                       raise ValueError('Cannot convert "%s" to isoformat' % repr(obj))
        return s

# -- Optional: if mx extensions are installed you may use mxDateTime ----
try:
    import mx.DateTime
    mxDateTime = True
except:
    mxDateTime = False
if mxDateTime:
    class mxDateTimeConverter(TimeConverter): # used optionally if installed
        def __init__(self):
            TimeConverter.__init__(self)
            self.types.add(type(mx.DateTime))
        def DateObjectFromCOMDate(self,comDate):
            return mx.DateTime.DateTimeFromCOMDate(comDate)
        def Date(self,year,month,day):
            return mx.DateTime.Date(year,month,day)
        def Time(self,hour,minute,second):
            return mx.DateTime.Time(hour,minute,second)
        def Timestamp(self,year,month,day,hour,minute,second):
            return mx.DateTime.Timestamp(year,month,day,hour,minute,second)
else:
    class mxDateTimeConverter(TimeConverter):
        pass    # if no mx is installed

class pythonDateTimeConverter(TimeConverter): # standard since Python 2.3
    def __init__(self):
        TimeConverter.__init__(self)
    def DateObjectFromCOMDate(self, comDate):
        if isinstance(comDate, datetime.datetime):
            odn = comDate.toordinal()
            tim = comDate.time()
            new = datetime.datetime.combine(datetime.datetime.fromordinal(odn), tim)
            return new
            # return comDate.replace(tzinfo=None) # make non aware
        elif isinstance(comDate, DateTime):
            fComDate = comDate.ToOADate() # ironPython clr Date/Time
        else:
            fComDate=float(comDate) #ComDate is number of days since 1899-12-31
        integerPart = int(fComDate)
        floatpart=fComDate-integerPart
        ##if floatpart == 0.0:
        ##    return datetime.date.fromordinal(integerPart + self._ordinal_1899_12_31)
        dte=datetime.datetime.fromordinal(integerPart + self._ordinal_1899_12_31) \
            + datetime.timedelta(milliseconds=floatpart*86400000)
        # millisecondsperday=86400000 # 24*60*60*1000
        return dte
    def Date(self,year,month,day):
        return datetime.date(year,month,day)
    def Time(self,hour,minute,second):
        return datetime.time(hour,minute,second)
    def Timestamp(self,year,month,day,hour,minute,second):
        return datetime.datetime(year,month,day,hour,minute,second)

class pythonTimeConverter(TimeConverter): # the old, ?nix type date and time
    def __init__(self): #caution: this Class gets confised by timezones and DST
        TimeConverter.__init__(self)
        self.types.add(time.struct_time)
    def DateObjectFromCOMDate(self,comDate):
        'Returns ticks since 1970'
        if isinstance(comDate,datetime.datetime):
            return comDate.timetuple()
        elif isinstance(comDate, DateTime): # ironPython clr date/time
            fcomDate = comDate.ToOADate()
        else:
            fcomDate = float(comDate)
        secondsperday=86400 # 24*60*60
        #ComDate is number of days since 1899-12-31, gmtime epoch is 1970-1-1 = 25569 days
        t=time.gmtime(secondsperday*(fcomDate-25569.0))
        return t  #year,month,day,hour,minute,second,weekday,julianday,daylightsaving=t
    def Date(self,year,month,day):
        return self.Timestamp(year,month,day,0,0,0)
    def Time(self,hour,minute,second):
        return time.gmtime((hour*60+minute)*60 + second)
    def Timestamp(self,year,month,day,hour,minute,second):
        return time.localtime(time.mktime((year,month,day,hour,minute,second,0,0,-1)))

base_dateconverter = pythonDateTimeConverter()

# ------ DB API required module attributes ---------------------
threadsafety=1  # TODO -- find out whether this module is actually BETTER than 1.

apilevel='2.0' #String constant stating the supported DB API level.

paramstyle='qmark' # the default parameter style

# ------ control for an extension which may become part of DB API 3.0 ---
accepted_paramstyles = ('qmark', 'named', 'format', 'pyformat', 'dynamic')

#------------------------------------------------------------------------------------------
# define similar types for generic conversion routines
adoIntegerTypes=(adc.adInteger,adc.adSmallInt,adc.adTinyInt,adc.adUnsignedInt,
                 adc.adUnsignedSmallInt,adc.adUnsignedTinyInt,
                 adc.adBoolean,adc.adError) #max 32 bits
adoRowIdTypes=(adc.adChapter,)          #v2.1 Rose
adoLongTypes=(adc.adBigInt,adc.adFileTime,adc.adUnsignedBigInt)
adoExactNumericTypes=(adc.adDecimal,adc.adNumeric,adc.adVarNumeric,adc.adCurrency)      #v2.3 Cole
adoApproximateNumericTypes=(adc.adDouble,adc.adSingle)                          #v2.1 Cole
adoStringTypes=(adc.adBSTR,adc.adChar,adc.adLongVarChar,adc.adLongVarWChar,
                adc.adVarChar,adc.adVarWChar,adc.adWChar)
adoBinaryTypes=(adc.adBinary,adc.adLongVarBinary,adc.adVarBinary)
adoDateTimeTypes=(adc.adDBTime, adc.adDBTimeStamp, adc.adDate, adc.adDBDate)
adoRemainingTypes=(adc.adEmpty,adc.adIDispatch,adc.adIUnknown,
                   adc.adPropVariant,adc.adArray,adc.adUserDefined,
                   adc.adVariant,adc.adGUID)

# this class is a trick to determine whether a type is a member of a related group of types. see PEP notes
class DBAPITypeObject(object):
    def __init__(self,valuesTuple):
        self.values = frozenset(valuesTuple)

    def __eq__(self,other):
        return other in self.values

    def __ne__(self, other):
        return other not in self.values

"""This type object is used to describe columns in a database that are string-based (e.g. CHAR). """
STRING   = DBAPITypeObject(adoStringTypes)

"""This type object is used to describe (long) binary columns in a database (e.g. LONG, RAW, BLOBs). """
BINARY   = DBAPITypeObject(adoBinaryTypes)

"""This type object is used to describe numeric columns in a database. """
NUMBER   = DBAPITypeObject(adoIntegerTypes + adoLongTypes + \
                           adoExactNumericTypes + adoApproximateNumericTypes)

"""This type object is used to describe date/time columns in a database. """

DATETIME = DBAPITypeObject(adoDateTimeTypes)
"""This type object is used to describe the "Row ID" column in a database. """
ROWID    = DBAPITypeObject(adoRowIdTypes)

OTHER = DBAPITypeObject(adoRemainingTypes)

# ------- utilities for translating python data types to ADO data types ---------------------------------
typeMap = { memoryViewType : adc.adVarBinary,
           float : adc.adDouble,
           type(None) : adc.adEmpty,
           str : adc.adBSTR,
           bool :adc.adBoolean,          #v2.1 Cole
           decimal.Decimal : adc.adDecimal,
           int: adc.adBigInt,
           bytes: adc.adVarBinary }

def pyTypeToADOType(d):
    tp=type(d)
    try:
        return typeMap[tp]
    except KeyError:  #   The type was not defined in the pre-computed Type table
        from . import dateconverter
        if tp in dateconverter.types:  # maybe it is one of our supported Date/Time types
            return adc.adDate
        #  otherwise, attempt to discern the type by probing the data object itself -- to handle duck typing
        if isinstance(d, StringTypes):
            return adc.adBSTR
        if isinstance(d, numbers.Integral):
            return adc.adBigInt
        if isinstance(d, numbers.Real):
            return adc.adDouble
        raise DataError('cannot convert "%s" (type=%s) to ADO'%(repr(d),tp))

# # # # # # # # # # # # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# functions to convert database values to Python objects
#------------------------------------------------------------------------
# variant type : function converting variant to Python value
def variantConvertDate(v):
    from . import dateconverter  # this function only called when adodbapi is running
    return dateconverter.DateObjectFromCOMDate(v)

def cvtString(variant):  # use to get old action of adodbapi v1 if desired
    if onIronPython:
        try:
            return variant.ToString()
        except:
            pass
    return str(variant)

def cvtDecimal(variant): #better name
    return _convertNumberWithCulture(variant, decimal.Decimal)

def cvtNumeric(variant): #older name - don't break old code
    return cvtDecimal(variant)

def cvtFloat(variant):
    return _convertNumberWithCulture(variant, float)

def _convertNumberWithCulture(variant, f):
    try:
        return f(variant)
    except (ValueError,TypeError,decimal.InvalidOperation):
        try:
            europeVsUS = str(variant).replace(",",".")
            return f(europeVsUS)
        except (ValueError,TypeError,decimal.InvalidOperation): pass

def cvtInt(variant):
    return int(variant)

def cvtLong(variant):  # only important in old versions where long and int differ
    return int(variant)

def cvtBuffer(variant):
    return bytes(variant)

def cvtUnicode(variant):
    return str(variant)

def identity(x): return x

def cvtUnusual(variant):
    if verbose > 1:
        sys.stderr.write('Conversion called for Unusual data=%s\n' % repr(variant))
    if isinstance(variant, DateTime):  # COMdate or System.Date
        from .adodbapi import dateconverter  # this will only be called when adodbapi is in use, and very rarely
        return dateconverter.DateObjectFromCOMDate(variant)
    return variant  # cannot find conversion function -- just give the data to the user

def convert_to_python(variant, func): # convert DB value into Python value
    if isinstance(variant, NullTypes):  # IronPython Null or None
        return None
    return func(variant)  # call the appropriate conversion function

class MultiMap(dict): #builds a dictionary from {(sequence,of,keys) : function}
    """A dictionary of ado.type : function -- but you can set multiple items by passing a sequence of keys"""
    #useful for defining conversion functions for groups of similar data types.
    def __init__(self, aDict):
        for k, v in list(aDict.items()):
            self[k] = v # we must call __setitem__
    def __setitem__(self, adoType, cvtFn):
        "set a single item, or a whole sequence of items"
        try: # user passed us a sequence, set them individually
            for type in adoType:
                dict.__setitem__(self, type, cvtFn)
        except TypeError: # a single value fails attempt to iterate
            dict.__setitem__(self, adoType, cvtFn)

#initialize variantConversions dictionary used to convert SQL to Python
# this is the dictionary of default conversion functions, built by the class above.
# this becomes a class attribute for the Connection, and that attribute is used
# to build the list of column conversion functions for the Cursor
variantConversions = MultiMap( {
    adoDateTimeTypes : variantConvertDate,
    adoApproximateNumericTypes: cvtFloat,
    adoExactNumericTypes: cvtDecimal, # use to force decimal rather than unicode
    adoLongTypes : cvtLong,
    adoIntegerTypes: cvtInt,
    adoRowIdTypes: cvtInt,
    adoStringTypes: identity,
    adoBinaryTypes: cvtBuffer,
    adoRemainingTypes: cvtUnusual })

# # # # # classes to emulate the result of cursor.fetchxxx() as a sequence of sequences # # # # #
    # "an ENUM of how my low level records are laid out"
RS_WIN_32, RS_ARRAY, RS_REMOTE = list(range(1,4))

class SQLrow(object): # a single database row
    # class to emulate a sequence, so that a column may be retrieved by either number or name
    def __init__(self, rows, index): # "rows" is an _SQLrows object, index is which row
        self.rows = rows # parent 'fetch' container object
        self.index = index # my row number within parent
    def __getattr__(self, name): # used for row.columnName type of value access
        try:
            return self._getValue(self.rows.columnNames[name.lower()])
        except KeyError:
            raise AttributeError('Unknown column name "{}"'.format(name))
    def _getValue(self,key):  # key must be an integer
        if self.rows.recordset_format == RS_ARRAY: # retrieve from two-dimensional array
            v = self.rows.ado_results[key,self.index]
        elif self.rows.recordset_format == RS_REMOTE:
            v = self.rows.ado_results[self.index][key]
        else:# pywin32 - retrieve from tuple of tuples
            v = self.rows.ado_results[key][self.index]
        if self.rows.converters is NotImplemented:
            return v
        return convert_to_python(v, self.rows.converters[key])

    def __len__(self):
        return self.rows.numberOfColumns

    def __getitem__(self,key):       # used for row[key] type of value access
        if isinstance(key,int):       # normal row[1] designation
            try:
                return self._getValue(key)
            except IndexError:
                raise
        if isinstance(key, slice):
            indices = key.indices(self.rows.numberOfColumns)
            vl = [self._getValue(i) for i in range(*indices)]
            return tuple(vl)
        try:
            return self._getValue(self.rows.columnNames[key.lower()])  # extension row[columnName] designation
        except (KeyError, TypeError):
            er, st, tr = sys.exc_info()
            raise er('No such key as "%s" in %s'%(repr(key),self.__repr__())).with_traceback(tr)
    def __iter__(self):
        return iter(self.__next__())
    def __next__(self):
        for n in range(self.rows.numberOfColumns):
            yield self._getValue(n)
    def __repr__(self): # create a human readable representation
        taglist = sorted(list(self.rows.columnNames.items()), key=lambda x: x[1])
        s = "<SQLrow={"
        for name, i in taglist:
            s += name + ':' + repr(self._getValue(i)) + ', '
        return s[:-2] + '}>'
    def __str__(self): # create a pretty human readable representation
        return str(tuple(str(self._getValue(i)) for i in range(self.rows.numberOfColumns)))

    # TO-DO implement pickling an SQLrow directly
    #def __getstate__(self): return self.__dict__
    #def __setstate__(self, d): self.__dict__.update(d)
    # which basically tell pickle to treat your class just like a normal one,
    # taking self.__dict__ as representing the whole of the instance state,
    #  despite the existence of the __getattr__.
    # # # #
class SQLrows(object):
    # class to emulate a sequence for multiple rows using a container object
    def __init__(self, ado_results, numberOfRows, cursor):
        self.ado_results = ado_results # raw result of SQL get
        try:
            self.recordset_format = cursor.recordset_format
            self.numberOfColumns = cursor.numberOfColumns
            self.converters = cursor.converters
            self.columnNames = cursor.columnNames
        except AttributeError:
            self.recordset_format = RS_ARRAY
            self.numberOfColumns = 0
            self.converters = []
            self.columnNames = {}
        self.numberOfRows = numberOfRows

    def __len__(self):
        return self.numberOfRows

    def __getitem__(self, item):     # used for row or row,column access
        if not self.ado_results:
            return []
        if isinstance(item, slice): # will return a list of row objects
            indices = item.indices(self.numberOfRows)
            return [SQLrow(self, k) for k in range(*indices)]
        elif isinstance(item, tuple) and len(item)==2:
            # d = some_rowsObject[i,j] will return a datum from a two-dimension address
            i, j = item
            if not isinstance(j, int):
                try:
                    j = self.columnNames[j.lower()] # convert named column to numeric
                except KeyError:
                    raise KeyError('adodbapi: no such column name as "%s"'%repr(j))
            if self.recordset_format == RS_ARRAY: # retrieve from two-dimensional array
                v = self.ado_results[j,i]
            elif self.recordset_format == RS_REMOTE:
                v = self.ado_results[i][j]
            else: # pywin32 - retrieve from tuple of tuples
                v = self.ado_results[j][i]
            if self.converters is NotImplemented:
                return v
            return convert_to_python(v, self.converters[j])
        else:
            row = SQLrow(self, item) # new row descriptor
            return row
    def __iter__(self):
        return iter(self.__next__())

    def __next__(self):
        for n in range(self.numberOfRows):
            row = SQLrow(self, n)
            yield row
            # # # # #


    # # # # # functions to re-format SQL requests to other paramstyle requirements # # # # # # # # # #
def changeNamedToQmark(op):  #convert from 'named' paramstyle to ADO required '?'mark parameters
    outOp = ''
    outparms=[]
    chunks = op.split("'")   #quote all literals -- odd numbered list results are literals.
    inQuotes = False
    for chunk in chunks:
        if inQuotes: # this is inside a quote
            if chunk == '': # double apostrophe to quote one apostrophe
                outOp = outOp[:-1]  # so take one away
            else:
                outOp += "'"+chunk+"'" # else pass the quoted string as is.
        else: # is SQL code -- look for a :namedParameter
            while chunk: # some SQL string remains
                sp = chunk.split(':',1)
                outOp += sp[0]  # concat the part up to the :
                s = ''
                try:
                    chunk = sp[1]
                except IndexError:
                    chunk = None
                if chunk:  # there was a parameter - parse it out
                    i = 0
                    c = chunk[0]
                    while c.isalnum() or c == '_':
                        i += 1
                        try:
                            c = chunk[i]
                        except IndexError:
                            break
                    s = chunk[:i]
                    chunk = chunk[i:]
                if s:
                    outparms.append(s) # list the parameters in order
                    outOp += '?'  # put in the Qmark
        inQuotes = not inQuotes
    return outOp, outparms

def changeFormatToQmark(op):  #convert from 'format' paramstyle to ADO required '?'mark parameters
    outOp = ''
    outparams = []
    chunks = op.split("'")   #quote all literals -- odd numbered list results are literals.
    inQuotes = False
    for chunk in chunks:
        if inQuotes:
            if outOp != '' and chunk=='': # he used a double apostrophe to quote one apostrophe
                outOp = outOp[:-1]  # so take one away
            else:
                outOp += "'"+chunk+"'" # else pass the quoted string as is.
        else:  # is SQL code -- look for a %s parameter
            if '%(' in chunk:  # ugh! pyformat!
                while chunk:  # some SQL string remains
                    sp = chunk.split('%(', 1)
                    outOp += sp[0]  # concat the part up to the %
                    if len(sp) > 1:
                        try:
                            s, chunk = sp[1].split(')s', 1)  # find the ')s'
                        except ValueError:
                            raise ProgrammingError('Pyformat SQL has incorrect format near "%s"' % chunk)
                        outparams.append(s)
                        outOp += '?'  # put in the Qmark
                    else:
                        chunk = None
            else:  # proper '%s' format
                sp = chunk.split('%s')  # make each %s
                outOp += "?".join(sp)   # into ?
        inQuotes =  not inQuotes # every other chunk is a quoted string
    return outOp, outparams



from statistics import mean
from treetopper._constants import (math,
                                   GRADE_NAMES,
                                   LOG_LENGTHS)
from treetopper._console_print import print_plot_logs


class Plot(object):
    """The Plot Class is an individual plot within a Stand's Inventory. It holds Timber Classes of the trees
       within the plot and runs calculations and statistics from those tree's metrics.

       For inventories, this class is meant to be added to the Stand Class using the Stand Class method of add_plot"""

    def __init__(self):
        self.trees = []
        self.tree_count = 0
        self.tpa = 0
        self.ba_ac = 0
        self.qmd = 0
        self.rd_ac = 0
        self.bf_ac = 0
        self.cf_ac = 0
        self.avg_hgt = 0
        self.vbar = 0
        self.metrics = ['tpa', 'ba_ac', 'rd_ac', 'bf_ac', 'cf_ac']

        self.species = {'totals_all': self._format_species_dict()}
        self.logs = {'totals_all': self._format_logs_species_dict()}

    def __getitem__(self, attribute: str):
        return self.__dict__[attribute]

    def add_tree(self, timber):
        """The timber argument should be one of the two Timber Classes (TimberQuick and/or TimberFull).
           The Timber Class is added to the plot's trees list and plot calculations and statistics are re-run"""
        self.trees.append(timber)
        self.tree_count += 1

        for met in self.metrics:
            setattr(self, met, getattr(self, met) + timber[met])

        self.qmd = math.sqrt((self.ba_ac / self.tpa) / .005454)
        self.vbar = self.bf_ac / self.ba_ac
        self.avg_hgt = mean([t.height for t in self.trees])
        self.hdr = mean([t.hdr for t in self.trees])

        self._update_species(timber)

        for l_num in timber.logs:
            log = timber.logs[l_num]
            if log.species not in self.logs:
                self.logs[log.species] = self._format_logs_species_dict()
            self._update_logs(log)

    def _format_logs_species_dict(self):
        """Returns a sub-dictionary for the species key of the plot's logs dictionary"""
        master = {}
        sub = {'lpa': 0,
               'bf_ac': 0,
               'cf_ac': 0}
        for grade in GRADE_NAMES:
            master[grade] = {rng: sub.copy() for rng in LOG_LENGTHS}
            master[grade].update({'totals_by_grade': sub.copy()})
            master[grade].update({'display': False})
        master['totals_by_length'] = {rng: sub.copy() for rng in LOG_LENGTHS}
        master['totals_by_length'].update({'totals_by_grade': sub.copy()})
        master['totals_by_length'].update({'display': False})
        return master

    def _update_logs(self, log):
        """Updated logs per acre, log board feet per acre, and log cubic feet per acre based on log grade and
           log length range for both the species key and the totals_all key of the plot's logs dictionary"""
        for key in [log.species, 'totals_all']:
            for sub in ['lpa', 'bf_ac', 'cf_ac']:
                self.logs[key][log.grade][log.length_range][sub] += log[sub]
                self.logs[key][log.grade]['totals_by_grade'][sub] += log[sub]
                self.logs[key][log.grade]['display'] = True
                self.logs[key]['totals_by_length'][log.length_range][sub] += log[sub]
                self.logs[key]['totals_by_length']['totals_by_grade'][sub] += log[sub]
                self.logs[key]['totals_by_length']['display'] = True

    def _format_species_dict(self):
        """Returns a sub-dictionary for the species key of the plot's species dictionary"""
        master = {'tpa': 0,
                  'ba_ac': 0,
                  'qmd': 0,
                  'avg_hgt': 0,
                  'hdr': 0,
                  'rd_ac': 0,
                  'bf_ac': 0,
                  'cf_ac': 0,
                  'vbar': 0}
        return master

    def _update_species(self, tree):
        """Updates the plot's overall calculations from the new data within the plot's species dictionary"""
        update_after = ['qmd', 'vbar', 'avg_hgt', 'hdr']
        if tree.species not in self.species:
            self.species[tree.species] = self._format_species_dict()

        for key in self.species[tree.species]:
            if key not in update_after:
                self.species[tree.species][key] += tree[key]
                self.species['totals_all'][key] += tree[key]

        for key in [tree.species, 'totals_all']:
            self.species[key]['qmd'] = math.sqrt((self.species[key]['ba_ac'] / self.species[key]['tpa']) / .005454)
            self.species[key]['vbar'] = self.species[key]['bf_ac'] / self.species[key]['ba_ac']
            if key == 'totals_all':
                self.species[key]['avg_hgt'] = mean([t.height for t in self.trees])
                self.species[key]['hdr'] = mean([t.hdr for t in self.trees])
            else:
                self.species[key]['avg_hgt'] = mean([t.height for t in self.trees if t.species == tree.species])
                self.species[key]['hdr'] = mean([t.hdr for t in self.trees if t.species == tree.species])





"""EXAMPLE OF AN INDIVIDUAL PLOT WITH RANDOMLY GENERATED TREES (TIMBERQUICK CLASS)"""

if __name__ == '__main__':
    from random import randrange
    from treetopper._testing import generate_random_trees_quick

    trees, tree_count = generate_random_trees_quick(randrange(4, 9))

    plot = Plot()
    for tree in trees:
        plot.add_tree(tree)

    print(f'Plot TreeCount: {len(trees)}')
    print(f'Plot TPA: {plot.tpa}')
    print(f'Plot BA_AC: {plot.ba_ac}')
    print(f'Plot BF_AC: {plot.bf_ac}')
    print(f'Plot AVG_HGT: {plot.avg_hgt}')
    print(f'Plot HDR: {plot.hdr}')
    print()

    for i, tree in enumerate(plot.trees):
        print(f'Tree #{i+1}')
        print(f'\tSpecies: {tree.species}')
        print(f'\tDBH: {tree.dbh}')
        print(f'\tHeight: {tree.height}')
        print(f'\tHDR: {tree.hdr}')
        print(f'\tBF_AC: {tree.bf_ac}')
        print(f'\tLOGS: {tree.logs}')
    print()

    print_plot_logs(plot)








import math
from os import startfile, getcwd
from os.path import join
from copy import deepcopy
from statistics import mean
from treetopper._console_print import print_thin_species
from treetopper._pdf_print import PDF
from treetopper._constants import extension_check


NOT_FULL_THIN_MESSAGE = """
** COULD NOT ACHIEVE TARGET DENSITY OF {target} {thin_param} **
** THINNING PARAMETERS ONLY ALLOWED {actual} {thin_param} OF THE {needed} {thin_param} NEEDED TO BE HARVESTED **
   THINNING PARAMETERS:
   \tSPECIES: {species}
   \tMIN DBH: {min_dbh}
   \tMAX DBH: {max_dbh}
"""

SUCCESS_MESSAGE = """
TARGET DENSITY OF {target} {thin_param} ACHIEVED
THINNING PARAMETERS:
   \tSPECIES: {species}
   \tMIN DBH: {min_dbh}
   \tMAX DBH: {max_dbh}
"""


class Thin(object):
    """The Thin Class is the parent class of the three thinning child classes: ThinTPA, ThinBA, and ThinRD.

       The thinning classes create a deepcopy of the stand class so as not to disrupt any of the tree information.

       The thinning classes start to thin the stand by getting a dictionary of whole-number diameters based on species,
       these contain per-diameter metrics such as tpa, ba_ac, rd_ac, bf_ac..., and these metrics, other than tpa,
       also have a corresponding metric per tree value, by dividing the per-diameter metric by the per-diameter tpa.

       The thinnings then calculate a harvest-ratio to lower the stand's density to the target density, in accordance to the
       species and diameter limitations (if any). The harvest ratio is used to calculate the removal and residual per-diameter TPAs,
       which are then used to calculate the other removal and residual per-diameter metrics by multiplying the respective TPAs by the
       per-diameter metric per tree values"""

    def __init__(self, stand, target_density: int, species_to_cut: list, min_dbh_to_cut: int, max_dbh_to_cut: int):
        self.stand = deepcopy(stand)
        self.target = target_density
        self.min_dbh = min_dbh_to_cut
        self.max_dbh = max_dbh_to_cut
        self.species = species_to_cut

        self.species_data = None
        self.target_metric = None

        self.full_thin = [True, None, None]

        self.keys = ['tpa', 'ba_ac', 'rd_ac', 'bf_ac', 'cf_ac']
        self.adds = [['ba_ac', 'ba_tree'], ['rd_ac', 'rd_tree'], ['bf_ac', 'bf_tree'], ['cf_ac', 'cf_tree']]
        self.conditions = ['current_', 'residual_', 'removal_']

        for key in self.keys:
            setattr(self, f'{self.conditions[0]}{key}', 0)

        self.trees = self._aggregate_trees()

    def __getitem__(self, attribute: str):
        return self.__dict__[attribute]

    def console_report(self):
        """Returns a console-formatted string of the stands conditions: current, removals, and residual shown by species and totals,
        to be called by the standard print() function"""
        species, min_dbh, max_dbh = self._get_message_params_report()
        thin_param = self.full_thin[2].replace('_', '/').upper()
        console_text = '\nTHINNING RESULTS\n'
        if not self.full_thin[0]:
            needed = round(self[f'current_{self.full_thin[2]}'] - self.target, 1)
            console_text += NOT_FULL_THIN_MESSAGE.format(target=self.target, thin_param=thin_param, actual=round(self.full_thin[1], 1),
                                                   needed=needed, species=', '.join(species), min_dbh=min_dbh, max_dbh=max_dbh)
            console_text += print_thin_species(self.species_data)
        else:
            console_text += SUCCESS_MESSAGE.format(target=self.target, thin_param=thin_param, species=', '.join(species),
                                              min_dbh=min_dbh, max_dbh=max_dbh)
            console_text += print_thin_species(self.species_data)
        # print(console_text)
        return console_text

    def pdf_report(self, filename: str, directory: str = None):
        """Exports a pdf of the thinning report to a user specified directory or if directory is None, to the current working directory"""
        if not self.full_thin[0]:
            needed = round(self[f'current_{self.full_thin[2]}'] - self.target)
            l1 = f"""** COULD NOT ACHIEVE TARGET DENSITY OF {self.target} {self.full_thin[2].replace('_', ' ').upper()} **"""
            l2 = f"""** THINNING PARAMETERS ONLY ALLOWED {round(self.full_thin[1], 1)} {self.full_thin[2].replace('_', '/').upper()} OF THE {needed} NEEDED TO BE HARVESTED **"""
            message = [l1, l2]
        else:
            message = None

        check = extension_check(filename, '.pdf')
        if directory:
            file = join(directory, check)
        else:
            file = join(getcwd(), check)
        pdf = PDF()
        pdf.alias_nb_pages()
        pdf.add_page()
        pdf.compile_thin_report(self, message=message)
        pdf.output(file, 'F')
        startfile(file)

    def _aggregate_trees(self):
        """Returns the compiled whole-number diameters dictionary separated by species"""
        cur = self.conditions[0]
        tree_dbh_data = {}
        for plot in self.stand.plots:
            for tree in plot.trees:
                d = int(tree.dbh)
                if tree.species not in tree_dbh_data:
                    tree_dbh_data[tree.species] = {}
                if d not in tree_dbh_data[tree.species]:
                    tree_dbh_data[tree.species][d] = {}
                    for condition in self.conditions:
                        tree_dbh_data[tree.species][d][condition] = {key: 0 for key in self.keys}
                    tree_dbh_data[tree.species][d].update({'trees': []})
                for key in self.keys:
                    tree_dbh_data[tree.species][d][cur][key] += tree[key] / self.stand.plot_count
                    setattr(self, f'{cur}{key}', getattr(self, f'{cur}{key}') + (tree[key] / self.stand.plot_count))
                tree_dbh_data[tree.species][d]['trees'].append(tree)

        for spp in tree_dbh_data:
            for d in tree_dbh_data[spp]:
                for add in self.adds:
                    tree_dbh_data[spp][d][cur][add[1]] = tree_dbh_data[spp][d][cur][add[0]] / tree_dbh_data[spp][d][cur]['tpa']
        return tree_dbh_data

    def _thin_to_target(self, target_metric):
        """Calculates the harvest ratio to produce the target density and then modifies the dictionary
           self.trees (which was created by self.aggregate_trees()) to the correct removal and residual values"""
        cur, res, rem = self.conditions

        if self.species == 'all':
            species = [spp for spp in self.trees]
        else:
            species = self.species
        sums = []
        for spp in self.trees:
            if spp in species:
                for d in self.trees[spp]:
                    if self.min_dbh <= d <= self.max_dbh:
                        sums.append(self.trees[spp][d][cur][target_metric])
        s = sum(sums)
        target_ratio = 1 - ((self[f'{cur}{target_metric}'] - self.target) / s)
        if target_ratio < 0:
            ratio = 0
            self.full_thin = [False, s, target_metric]
        else:
            ratio = target_ratio
            self.full_thin = [True, s, target_metric]

        for spp in self.trees:
            if spp in species:
                for d in self.trees[spp]:
                    if self.min_dbh <= d <= self.max_dbh:
                        self.trees[spp][d][res]['tpa'] = ratio * self.trees[spp][d][cur]['tpa']
                        self.trees[spp][d][rem]['tpa'] = (1 - ratio) * self.trees[spp][d][cur]['tpa']
                        for condition in self.conditions[1:]:
                            for add in self.adds:
                                self.trees[spp][d][condition][add[0]] = self.trees[spp][d][cur][add[1]] * self.trees[spp][d][condition]['tpa']
                    else:
                        for add in self.adds:
                            self.trees[spp][d][res][add[0]] = self.trees[spp][d][cur][add[0]]
                        self.trees[spp][d][res]['tpa'] = self.trees[spp][d][cur]['tpa']
            else:
                for d in self.trees[spp]:
                    for add in self.adds:
                        self.trees[spp][d][res][add[0]] = self.trees[spp][d][cur][add[0]]
                    self.trees[spp][d][res]['tpa'] = self.trees[spp][d][cur]['tpa']

    def _get_species_conditions(self):
        """Returns an dictionary composed of the aggregate of the species totaled from the per-diameter values of the
           resultant self.trees (modified from self.thin_to_target())"""
        master = {}
        for condition in self.conditions:
            total_trees = []
            master[condition] = {'totals_all': {key: 0 for key in self.keys}}
            for spp in self.trees:
                trees = []
                master[condition][spp] = {key: 0 for key in self.keys}
                for d in self.trees[spp]:
                    for tree in self.trees[spp][d]['trees']:
                        if condition == 'current_':
                            trees.append(tree)
                            total_trees.append(tree)
                        elif condition == 'residual_':
                            d_rng = [d for d in self.trees[spp]]
                            min_ = min(d_rng)
                            max_ = max(d_rng)
                            if self.min_dbh <= min_ and self.max_dbh >= max_:
                                trees.append(tree)
                                total_trees.append(tree)
                            else:
                                if d < self.min_dbh or d > self.max_dbh:
                                    trees.append(tree)
                                    total_trees.append(tree)
                        else:
                            if self.min_dbh <= d <= self.max_dbh:
                                trees.append(tree)
                                total_trees.append(tree)
                    for key in self.keys:
                        master[condition][spp][key] += self.trees[spp][d][condition][key]
                        master[condition]['totals_all'][key] += self.trees[spp][d][condition][key]

                if master[condition][spp]['tpa'] > 0:
                    self._update_species_conditions_dict(master[condition][spp], trees)
                else:
                    del master[condition][spp]
            self._update_species_conditions_dict(master[condition]['totals_all'], total_trees)
        return master

    def _update_species_conditions_dict(self, master_condition_species, trees):
        """Updates the species data dictionary being compiled in self._get_species_conditions(), used internally"""
        master_condition_species.update({'qmd': math.sqrt((master_condition_species['ba_ac'] / master_condition_species['tpa']) / 0.005454)})
        master_condition_species.update({'vbar': master_condition_species['bf_ac'] / master_condition_species['ba_ac']})
        master_condition_species.update({'avg_hgt': mean([tree.height for tree in trees])})
        master_condition_species.update({'hdr': mean([tree.hdr for tree in trees])})

    def _get_message_params_report(self):
        """Returns the thinning parameters, formatted to be displayed in the console report or print report"""
        if self.species == 'all':
            species = [spp for spp in self.trees]
        else:
            species = self.species

        if self.min_dbh == 0:
            min_dbh = 'No Minimum'
        else:
            min_dbh = f'{self.min_dbh} inches'

        if self.max_dbh == 999:
            max_dbh = 'No Maximum'
        else:
            max_dbh = f'{self.max_dbh} inches'
        return species, min_dbh, max_dbh

    def _check_density(self):
        """Will check to see if the target density is in fact lower than the total stand density. If the target density is higher,
           it will throw a custom exception: TargetDensityError"""
        current_density = self[f'current_{self.target_metric}']
        met = self.target_metric.replace('_', '/').upper()
        if current_density < self.target:
            raise TargetDenistyError(self.target, current_density, met)


class ThinTPA(Thin):
    """Thin by Trees per Acre"""

    def __init__(self, stand, target_density: int, species_to_cut: list = 'all', min_dbh_to_cut: int = 0, max_dbh_to_cut: int = 999):
        super(ThinTPA, self).__init__(stand, target_density, species_to_cut, min_dbh_to_cut, max_dbh_to_cut)
        self.target_metric = 'tpa'
        self._check_density()
        self._thin_to_target(self.target_metric)
        self.species_data = self._get_species_conditions()


class ThinBA(Thin):
    """Thin by Basal Area per Acre"""

    def __init__(self, stand, target_density: int, species_to_cut: list = 'all', min_dbh_to_cut: int = 0, max_dbh_to_cut: int = 999):
        super(ThinBA, self).__init__(stand, target_density, species_to_cut, min_dbh_to_cut, max_dbh_to_cut)
        self.target_metric = 'ba_ac'
        self._check_density()
        self._thin_to_target(self.target_metric)
        self.species_data = self._get_species_conditions()


class ThinRD(Thin):
    """Thin by Relative Density per Acre"""

    def __init__(self, stand, target_density: int, species_to_cut: list = 'all', min_dbh_to_cut: int = 0, max_dbh_to_cut: int = 999):
        super(ThinRD, self).__init__(stand, target_density, species_to_cut, min_dbh_to_cut, max_dbh_to_cut)
        self.target_metric = 'rd_ac'
        self._check_density()
        self._thin_to_target(self.target_metric)
        self.species_data = self._get_species_conditions()


class TargetDenistyError(Exception):
    def __init__(self, target_density, current_density, target_metric):
        p1 = f"""Target Density of {target_density} {target_metric} """
        p2 = f"""is greater than Stand Total of {round(current_density, 1)} {target_metric}. """
        p3 = f"""Please lower Target Density"""
        self.message = p1 + p2 + p3
        super(TargetDenistyError, self).__init__(self.message)



if __name__ == '__main__':
    from treetopper.stand import Stand

    stand = Stand('OK2', 46.94)
    stand.from_csv_full('../example_csv_and_xlsx/Example_CSV_full.csv')

    thin = ThinBA(stand, 160)
    thin.console_report()












import sys

import win32evtlog


def main():
    path = 'System'
    num_events = 5
    if len(sys.argv) > 2:
        path = sys.argv[1]
        num_events = int(sys.argv[2])
    elif len(sys.argv) > 1:
        path = sys.argv[1]

    query = win32evtlog.EvtQuery(path, win32evtlog.EvtQueryForwardDirection)
    events = win32evtlog.EvtNext(query, num_events)
    context = win32evtlog.EvtCreateRenderContext(win32evtlog.EvtRenderContextSystem)

    for i, event in enumerate(events, 1):
        result = win32evtlog.EvtRender(event, win32evtlog.EvtRenderEventValues, Context=context)

        print('Event {}'.format(i))

        level_value, level_variant = result[win32evtlog.EvtSystemLevel]
        if level_variant != win32evtlog.EvtVarTypeNull:
            if level_value == 1:
                print('    Level: CRITICAL')
            elif level_value == 2:
                print('    Level: ERROR')
            elif level_value == 3:
                print('    Level: WARNING')
            elif level_value == 4:
                print('    Level: INFO')
            elif level_value == 5:
                print('    Level: VERBOSE')
            else:
                print('    Level: UNKNOWN')

        time_created_value, time_created_variant = result[win32evtlog.EvtSystemTimeCreated]
        if time_created_variant != win32evtlog.EvtVarTypeNull:
            print('    Timestamp: {}'.format(time_created_value.isoformat()))

        computer_value, computer_variant = result[win32evtlog.EvtSystemComputer]
        if computer_variant != win32evtlog.EvtVarTypeNull:
            print('    FQDN: {}'.format(computer_value))

        provider_name_value, provider_name_variant = result[win32evtlog.EvtSystemProviderName]
        if provider_name_variant != win32evtlog.EvtVarTypeNull:
            print('    Provider: {}'.format(provider_name_value))

            try:
                metadata = win32evtlog.EvtOpenPublisherMetadata(provider_name_value)
            # pywintypes.error: (2, 'EvtOpenPublisherMetadata', 'The system cannot find the file specified.')
            except Exception:
                pass
            else:
                try:
                    message = win32evtlog.EvtFormatMessage(metadata, event, win32evtlog.EvtFormatMessageEvent)
                # pywintypes.error: (15027, 'EvtFormatMessage: allocated 0, need buffer of size 0', 'The message resource is present but the message was not found in the message table.')
                except Exception:
                    pass
                else:
                    print('    Message: {}'.format(message))


if __name__=='__main__':
    main()



import setuptools

keywords = ['forestry', 'natural resources', 'natural', 'resources',
            'forest', 'environmental', 'environmental science', 'science',
            'timber', 'tree', 'trees', 'west', 'coast', 'west coast', 'logging',
            'cruising', 'scaling', 'inventory', 'forests', 'board feet', 'cubic feet',
            'DBH', 'diameter at breast height', 'DIB', 'diameter inside bark', 'RD',
            'relative density', 'HDR', 'height to diamater ratio', 'species', 'VBAR',
            'volume to basal area ratio', 'BA', 'basal area', 'TPA', 'trees per acre',
            'scribner', 'timber marketing', 'forest marketing', 'timber management',
            'forest management']

with open("README.txt", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name = 'treetopper',
    version = "1.0.8",
    author = 'Zach Beebe',
    author_email = 'z.beebe@yahoo.com',
    description = 'Python module for calculating Timber data from tree species of the west coast',
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = 'https://github.com/zacharybeebe/treetopper',
    license = 'MIT',
    packages = setuptools.find_packages(),
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"],
    keywords = keywords,
    python_requires = '>=3.6',
    py_modules = ['treetopper'],
    install_requires=[
        'fpdf >= 1.7.2',
        'openpyxl >= 3.0.6',
        'pyodbc == 4.0.30',
        "pywin32 >=1.0 ; sys_platform == 'win32'"
    ]
)



remote = False  # automatic testing of remote access has been removed here

def try_connection(verbose, *args, **kwargs):
    import adodbapi

    dbconnect = adodbapi.connect
    try:
        s = dbconnect(*args, **kwargs) # connect to server
        if verbose:
            print('Connected to:', s.connection_string)
            print('which has tables:', s.get_table_names())
        s.close()  # thanks, it worked, goodbye
    except adodbapi.DatabaseError as inst:
        print(inst.args[0])    # should be the error message
        print('***Failed getting connection using=',repr(args),repr(kwargs))
        return False, (args, kwargs), None

    print("  (successful)")

    return True, (args, kwargs, remote), dbconnect


def try_operation_with_expected_exception(expected_exception_list, some_function, *args, **kwargs):
    try:
        some_function(*args, **kwargs)
    except expected_exception_list as e:
        return True, e
    except:
        raise  # an exception other than the expected occurred
    return False, 'The expected exception did not occur'



"""
This demonstrates the creation of miniversions of a file during a transaction.
The FSCTL_TXFS_CREATE_MINIVERSION control code saves any changes to a new
miniversion (effectively a savepoint within a transaction).
"""

import win32file, win32api, win32transaction, winerror
import win32con, winioctlcon
import struct
import os
from pywin32_testutil import str2bytes # py3k-friendly helper

def demo():
    """
    Definition of buffer used with FSCTL_TXFS_CREATE_MINIVERSION:
    typedef struct _TXFS_CREATE_MINIVERSION_INFO{
        USHORT StructureVersion;
        USHORT StructureLength;
        ULONG BaseVersion;
        USHORT MiniVersion;}
    """
    buf_fmt='HHLH0L'   ## buffer size must include struct padding
    buf_size=struct.calcsize(buf_fmt)

    tempdir=win32api.GetTempPath()
    tempfile=win32api.GetTempFileName(tempdir,'cft')[0]
    print("Demonstrating transactions on tempfile", tempfile)
    f=open(tempfile,'w')
    f.write('This is original file.\n')
    f.close()

    trans=win32transaction.CreateTransaction(Description='Test creating miniversions of a file')
    hfile=win32file.CreateFileW(tempfile, win32con.GENERIC_READ|win32con.GENERIC_WRITE,
        win32con.FILE_SHARE_READ|win32con.FILE_SHARE_WRITE,
        None, win32con.OPEN_EXISTING, 0 , None, Transaction=trans)

    win32file.WriteFile(hfile, str2bytes('This is first miniversion.\n'))
    buf=win32file.DeviceIoControl(hfile, winioctlcon.FSCTL_TXFS_CREATE_MINIVERSION,None,buf_size,None)
    struct_ver, struct_len, base_ver, ver_1=struct.unpack(buf_fmt, buf)

    win32file.SetFilePointer(hfile, 0, win32con.FILE_BEGIN)
    win32file.WriteFile(hfile, str2bytes('This is second miniversion!\n'))
    buf=win32file.DeviceIoControl(hfile, winioctlcon.FSCTL_TXFS_CREATE_MINIVERSION,None,buf_size,None)
    struct_ver, struct_len, base_ver, ver_2=struct.unpack(buf_fmt, buf)
    hfile.Close()

    ## miniversions can't be opened with write access
    hfile_0=win32file.CreateFileW(tempfile, win32con.GENERIC_READ,
        win32con.FILE_SHARE_READ|win32con.FILE_SHARE_WRITE,
        None, win32con.OPEN_EXISTING, 0 , None, Transaction=trans, MiniVersion=base_ver)
    print('version:',base_ver,win32file.ReadFile(hfile_0, 100))
    hfile_0.Close()

    hfile_1=win32file.CreateFileW(tempfile, win32con.GENERIC_READ,
        win32con.FILE_SHARE_READ|win32con.FILE_SHARE_WRITE,
        None, win32con.OPEN_EXISTING, 0 , None, Transaction=trans, MiniVersion=ver_1)
    print('version:',ver_1,win32file.ReadFile(hfile_1, 100))
    hfile_1.Close()

    hfile_2=win32file.CreateFileW(tempfile, win32con.GENERIC_READ,
        win32con.FILE_SHARE_READ|win32con.FILE_SHARE_WRITE,
        None, win32con.OPEN_EXISTING, 0 , None, Transaction=trans, MiniVersion=ver_2)
    print('version:',ver_2,win32file.ReadFile(hfile_2, 100))
    hfile_2.Close()

    ## MiniVersions are destroyed when transaction is committed or rolled back
    win32transaction.CommitTransaction(trans)

    os.unlink(tempfile)

if __name__ == "__main__":
    # When run on CI, this fails with NOT_SUPPORTED, so don't have that cause "failure"
    try:
        demo()
    except win32file.error as e:
        if e.winerror == winerror.ERROR_NOT_SUPPORTED:
            print("These features are not supported by this filesystem.")
        else:
            raise


